#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import parl
import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.distributions import Normal

__all__ = ['PPO']


class PPO(parl.Algorithm):
    def __init__(self,
                 model,
                 clip_param,
                 value_loss_coef,
                 entropy_coef,
                 initial_lr,
                 eps=None,
                 max_grad_norm=None,
                 use_clipped_value_loss=True):
        self.model = model

        self.clip_param = clip_param

        self.value_loss_coef = value_loss_coef
        self.entropy_coef = entropy_coef

        self.max_grad_norm = max_grad_norm
        self.use_clipped_value_loss = use_clipped_value_loss

        self.optimizer = optim.Adam(model.parameters(), lr=initial_lr, eps=eps)

    def learn(self, obs_batch, actions_batch, value_preds_batch, return_batch,
              old_action_log_probs_batch, adv_targ):
        values = self.model.value(obs_batch)
        mean, log_std = self.model.policy(obs_batch)
        dist = Normal(mean, log_std.exp())

        action_log_probs = dist.log_prob(actions_batch).sum(-1, keepdim=True)
        dist_entropy = dist.entropy().sum(-1).mean()

        ratio = torch.exp(action_log_probs - old_action_log_probs_batch)
        surr1 = ratio * adv_targ
        surr2 = torch.clamp(ratio, 1.0 - self.clip_param,
                            1.0 + self.clip_param) * adv_targ
        action_loss = -torch.min(surr1, surr2).mean()

        if self.use_clipped_value_loss:
            value_pred_clipped = value_preds_batch + \
                (values - value_preds_batch).clamp(-self.clip_param, self.clip_param)
            value_losses = (values - return_batch).pow(2)
            value_losses_clipped = (value_pred_clipped - return_batch).pow(2)
            value_loss = 0.5 * torch.max(value_losses,
                                         value_losses_clipped).mean()
        else:
            value_loss = 0.5 * (return_batch - values).pow(2).mean()

        self.optimizer.zero_grad()
        (value_loss * self.value_loss_coef + action_loss -
         dist_entropy * self.entropy_coef).backward()
        nn.utils.clip_grad_norm_(self.model.parameters(), self.max_grad_norm)
        self.optimizer.step()

        return value_loss.item(), action_loss.item(), dist_entropy.item()

    def sample(self, obs):
        value = self.model.value(obs)
        mean, log_std = self.model.policy(obs)
        dist = Normal(mean, log_std.exp())
        action = dist.sample()
        action_log_probs = dist.log_prob(action).sum(-1, keepdim=True)

        return value, action, action_log_probs

    def predict(self, obs):
        mean, _ = self.model.policy(obs)
        return mean

    def value(self, obs):
        return self.model.value(obs)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import threading
from parl.remote import Master, Worker


def main(args):
    """Start a master or a worker through:

    1. xparl start --port 1234
    2. xparl connect --address localhost:1234 --cpu_num 8

    """

    if args.name == 'master':
        port = args.port
        monitor_port = args.monitor_port
        master = Master(port, monitor_port)
        master.run()

    elif args.name == 'worker':
        address = args.address
        log_server_port = args.log_server_port
        cpu_num = int(args.cpu_num) if args.cpu_num else None
        worker = Worker(address, cpu_num, log_server_port)
        worker.run()

    else:
        raise NotImplementedError


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--name', default='master', type=str, help='master/worker')
    parser.add_argument('--port', default='1234', type=str)
    parser.add_argument('--address', default='localhost:1234', type=str)
    parser.add_argument('--cpu_num', default='', type=str)
    parser.add_argument('--monitor_port', default='', type=str)
    parser.add_argument('--log_server_port', default='', type=str)
    args = parser.parse_args()
    main(args)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from six.moves import queue
import time
import threading

from collections import namedtuple
from parl.remote.exceptions import FutureFunctionError
from parl.remote.future_mode.future_object import FutureObject

XPARL_RESERVED_PREFIX = "_xparl"
RESERVED_NAME_ERROR_STR = "Name starting with `_xparl` is the reserved name in xparl, please don't use the name `{}`."

CallingRequest = namedtuple(
    'CallingRequest',
    ['calling_type', 'attr', 'value', 'args', 'kwargs', 'future_return_queue'])


def proxy_wrapper_nowait_func(remote_wrapper):
    '''This function implements the remote decorator for asynchronous mode.
    With this decorator, the function called by user return a `future` object immediately
    and it will not be blocked. Users can get the real return of the function by calling 
    the `get` function of the `future` object.

    For example:
        ```python
        import parl

        @parl.remote_class(wait=False)
        class Actor(object):
            def __init__(self):
                pass

            def func(self):
                # do something
                return 0

        parl.connect("localhost:8010")

        actors = [Actor() for _ in range(10)]
        future_objects = [actor.func() for actor in actors]
        results = [future_obj.get() for future_obj in future_objects]
        ```
    '''
    original_class = remote_wrapper._original
    max_memory = remote_wrapper._max_memory

    class ProxyWrapperNoWait(object):
        def __init__(self, *args, **kwargs):
            for key in kwargs:
                assert not key.startswith(
                    XPARL_RESERVED_PREFIX), RESERVED_NAME_ERROR_STR.format(key)

            kwargs['_xparl_proxy_wrapper_nowait'] = self
            kwargs['_xparl_remote_class'] = original_class
            kwargs['_xparl_remote_class_max_memory'] = max_memory

            self._xparl_remote_wrapper_calling_queue = queue.Queue()
            self._xparl_remote_wrapper_internal_lock = threading.Lock()
            self._xparl_calling_finished_event = threading.Event()
            self._xparl_remote_object_exception = None

            object_thread = threading.Thread(
                target=self._run_object_in_backend, args=(args, kwargs))
            object_thread.setDaemon(True)
            object_thread.start()

        def _run_object_in_backend(self, args, kwargs):
            try:
                self._xparl_remote_wrapper_obj = remote_wrapper(
                    *args, **kwargs)
                for key in self._xparl_remote_wrapper_obj.get_attrs():
                    assert not key.startswith(
                        XPARL_RESERVED_PREFIX), RESERVED_NAME_ERROR_STR.format(
                            key)
            except Exception as e:
                async_error = FutureFunctionError('__init__')
                self._xparl_remote_object_exception = async_error
                self._xparl_calling_finished_event.set()
                raise e
            """
            NOTE:
                We should set the event after the initialization of self._xparl_remote_wrapper_obj.
                Only after the initialization is complete can we call the function of actor.
            """
            self._xparl_calling_finished_event.set()

            while True:
                calling_request = self._xparl_remote_wrapper_calling_queue.get(
                )
                assert calling_request.calling_type in ["setattr", "getattr"]

                try:
                    if calling_request.calling_type == "setattr":
                        self._xparl_remote_wrapper_obj.set_remote_attr(
                            calling_request.attr, calling_request.value)
                    else:  # getattr
                        is_attribute = self._xparl_remote_wrapper_obj.has_attr(
                            calling_request.attr)

                        if is_attribute:
                            return_result = self._xparl_remote_wrapper_obj.get_remote_attr(
                                calling_request.attr)
                        else:
                            function_wrapper = self._xparl_remote_wrapper_obj.get_remote_attr(
                                calling_request.attr)
                            return_result = function_wrapper(
                                *calling_request.args,
                                **calling_request.kwargs)
                        calling_request.future_return_queue.put(return_result)

                except Exception as e:
                    async_error = FutureFunctionError(calling_request.attr)
                    self._xparl_remote_object_exception = async_error
                    if calling_request.calling_type == "getattr":
                        calling_request.future_return_queue.put(async_error)
                    raise e

                finally:
                    self._xparl_calling_finished_event.set()

        def __getattr__(self, attr):
            self._xparl_remote_wrapper_internal_lock.acquire()

            self._xparl_calling_finished_event.wait(
            )  # waiting for last function finishing before calling has_attr

            if self._xparl_remote_object_exception is not None:
                time.sleep(
                    0.1
                )  # waiting for another thread printing the error message
                raise self._xparl_remote_object_exception
            """
            Don't use the following way, which will call the __getattr__ function and acquire the lock again.
                is_attribute = self._xparl_remote_wrapper_obj.has_attr(attr)
            """
            is_attribute = self.__dict__['_xparl_remote_wrapper_obj'].has_attr(
                attr)

            self._xparl_remote_wrapper_internal_lock.release()

            def wrapper(*args, **kwargs):
                self._xparl_remote_wrapper_internal_lock.acquire()

                self._xparl_calling_finished_event.wait()
                self._xparl_calling_finished_event.clear()

                if self._xparl_remote_object_exception is not None:
                    time.sleep(
                        0.1
                    )  # waiting for another thread printing the error message
                    raise self._xparl_remote_object_exception

                future_return_queue = queue.Queue()
                calling_request = CallingRequest(
                    calling_type="getattr",
                    attr=attr,
                    value=None,
                    args=args,
                    kwargs=kwargs,
                    future_return_queue=future_return_queue)

                self.__dict__['_xparl_remote_wrapper_calling_queue'].put(
                    calling_request)

                future_object = FutureObject(future_return_queue)

                self._xparl_remote_wrapper_internal_lock.release()
                return future_object

            if is_attribute:
                future_object = wrapper()
                return future_object.get()
            else:
                return wrapper

        def __setattr__(self, attr, value):
            if attr.startswith(XPARL_RESERVED_PREFIX):
                super(ProxyWrapperNoWait, self).__setattr__(attr, value)
            else:
                self._xparl_remote_wrapper_internal_lock.acquire()

                self._xparl_calling_finished_event.wait()
                self._xparl_calling_finished_event.clear()

                if self._xparl_remote_object_exception is not None:
                    time.sleep(
                        0.1
                    )  # waiting for another thread to print the error message
                    raise self._xparl_remote_object_exception

                calling_request = CallingRequest(
                    calling_type="setattr",
                    attr=attr,
                    value=value,
                    args=None,
                    kwargs=None,
                    future_return_queue=None)
                self._xparl_remote_wrapper_calling_queue.put(calling_request)

                self._xparl_remote_wrapper_internal_lock.release()

    return ProxyWrapperNoWait



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cloudpickle
import subprocess
import os
from parl.utils import SerializeError, DeserializeError

__all__ = ['dumps_argument', 'loads_argument', 'dumps_return', 'loads_return']

try:
    import pyarrow
    pyarrow_installed = True
except ImportError:
    pyarrow_installed = False

if pyarrow_installed:
    # Reference: https://github.com/apache/arrow/blob/f88474c84e7f02e226eb4cc32afef5e2bbc6e5b4/python/pyarrow/tests/test_serialization.py#L658-L682
    def _serialize_serializable(obj):
        return {"type": type(obj), "data": obj.__dict__}

    def _deserialize_serializable(obj):
        val = obj["type"].__new__(obj["type"])
        val.__dict__.update(obj["data"])
        return val

    context = pyarrow.default_serialization_context()

    # support deserialize in another environment
    context.set_pickle(cloudpickle.dumps, cloudpickle.loads)

    # support serialize and deserialize custom class
    context.register_type(
        object,
        "object",
        custom_serializer=_serialize_serializable,
        custom_deserializer=_deserialize_serializable)

    # if pyarrow is installed, parl will use pyarrow to serialize/deserialize objects.
    serialize = lambda data: pyarrow.serialize(data, context=context).to_buffer()
    deserialize = lambda data: pyarrow.deserialize(data, context=context)
else:
    # if pyarrow is not installed, parl will use cloudpickle to serialize/deserialize objects.
    serialize = lambda data: cloudpickle.dumps(data)
    deserialize = lambda data: cloudpickle.loads(data)


def dumps_argument(*args, **kwargs):
    """

    Serialize arguments passed to a function.

    args: 
        *args, **kwargs are general a commonly used representation of arguments in python.

    Returns:
        Implementation-dependent object in bytes.
    """
    try:
        ret = serialize([args, kwargs])
    except Exception as e:
        raise SerializeError(e)

    return ret


def loads_argument(data):
    """
    Restore bytes data to their initial data formats.

    Args:
        data: the output of `dumps_argument`.

    Returns:
        deserialized arguments [args, kwargs]
        like the input of `dumps_argument`, args is a tuple, and kwargs is a dict 
    """
    try:
        ret = deserialize(data)
    except Exception as e:
        raise DeserializeError(e)

    return ret


def dumps_return(data):
    """
    Serialize the return data of a function.

    Args:
        data: the output of a function.

    Returns:
        Implementation-dependent object in bytes.
    """
    try:
        ret = serialize(data)
    except Exception as e:
        raise SerializeError(e)

    return ret


def loads_return(data):
    """
    Deserialize the data generated by `dumps_return`.

    Args:
        data: the output of `dumps_return`

    Returns:
        deserialized data
    """
    try:
        ret = deserialize(data)
    except Exception as e:
        raise DeserializeError(e)

    return ret



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Fix cloudpickle compatible problem we known.
import compatible_trick

import os
os.environ['XPARL'] = 'True'
os.environ['CUDA_VISIBLE_DEVICES'] = ''
import argparse
import cloudpickle
import pickle
import psutil
import re
import sys
import tempfile
import threading
import time
import traceback
import zmq
from multiprocessing import Process, Pipe
from parl.utils import to_str, to_byte, get_ip_address, logger
from parl.remote.communication import loads_argument, loads_return,\
    dumps_argument, dumps_return
from parl.remote import remote_constants
from parl.utils.exceptions import SerializeError, DeserializeError
from parl.remote.message import InitializedJob
from parl.remote.utils import load_remote_class, redirect_output_to_file
from parl.remote.zmq_utils import create_server_socket, create_client_socket
from parl.remote.grpc_heartbeat import HeartbeatServerThread


class Job(object):
    """Base class for the job.

    After establishing connection with the remote object, the job will
    create a remote class instance locally and enter an infinite loop
    in a separate process, waiting for commands from the remote object.

    """

    def __init__(self, worker_address, log_server_address):
        """
        Args:
            worker_address(str): worker_address for sending job information(e.g, pid)

        Attributes:
            pid (int): Job process ID.
            max_memory (float): Maximum memory (MB) can be used by each remote instance.
        """
        self.max_memory = None

        self.job_address_receiver, job_address_sender = Pipe()
        self.job_id_receiver, job_id_sender = Pipe()

        self.worker_address = worker_address
        self.log_server_address = log_server_address
        self.job_ip = get_ip_address()
        self.pid = os.getpid()
        """
        NOTE:
            In Windows, it will raise errors when creating threading.Lock before starting multiprocess.Process.
        """
        self.lock = threading.Lock()
        th = threading.Thread(target=self._create_sockets)
        th.setDaemon(True)
        th.start()

        process = psutil.Process(self.pid)
        self.init_memory = float(process.memory_info()[0]) / (1024**2)

        self.run(job_address_sender, job_id_sender)

        with self.lock:
            self.remove_job_socket.send_multipart(
                [remote_constants.KILLJOB_TAG,
                 to_byte(self.job_address)])
            try:
                _ = self.remove_job_socket.recv_multipart()
            except zmq.error.Again as e:
                pass
            os._exit(0)

    def _create_sockets(self):
        """Create three sockets for each job in the main process.

        (1) job_socket(functional socket): sends job_address and heartbeat_address to worker.
        (2) reply_client_ping_socket: replies ping message of client.
        (3) remove_job_socket: sends a command to the corresponding worker to remove the job.
                               Used to ask the worker removing the dead job immediately 
                               instead of waiting for the heartbeat failure.

        Create two heartbeat server threads for each job:
        (1) worker_heartbeat_server_thread: reply heartbeat signal from the worker.
        (2) client_heartbeat_server_thread: reply heartbeat signal from the client.
        """
        # wait for another process to create reply socket
        self.job_address = self.job_address_receiver.recv()
        self.job_id = self.job_id_receiver.recv()

        self.ctx = zmq.Context()
        # create the job_socket
        self.job_socket = create_client_socket(
            self.ctx, self.worker_address, heartbeat_timeout=True)

        # a thread that reply ping signals from the client
        reply_client_ping_socket, port = create_server_socket(self.ctx)
        reply_client_ping_address = "{}:{}".format(self.job_ip, port)
        ping_thread = threading.Thread(
            target=self._reply_ping, args=(reply_client_ping_socket, ))
        ping_thread.setDaemon(True)
        ping_thread.start()

        # a thread that reply heartbeat signals from the worker
        def worker_heartbeat_exit_callback_func():
            logger.warning("[Job]lost connection with the worker, will exit")
            os._exit(1)

        worker_heartbeat_server_thread = HeartbeatServerThread(
            heartbeat_exit_callback_func=worker_heartbeat_exit_callback_func)
        worker_heartbeat_server_thread.setDaemon(True)
        worker_heartbeat_server_thread.start()

        # This function will be called only after the heartbeat server thread is started
        def client_heartbeat_exit_callback_func():
            with self.lock:
                self.remove_job_socket.send_multipart(
                    [remote_constants.KILLJOB_TAG,
                     to_byte(self.job_address)])
                try:
                    _ = self.remove_job_socket.recv_multipart()
                except zmq.error.Again as e:
                    pass
            logger.warning("[Job]lost connection with the client, will exit")
            os._exit(1)

        # a thread that reply heartbeat signals from the client
        self.client_heartbeat_server_thread = HeartbeatServerThread(
            heartbeat_exit_callback_func=client_heartbeat_exit_callback_func)
        self.client_heartbeat_server_thread.setDaemon(True)

        # sends job information to the worker
        initialized_job = InitializedJob(
            self.job_address, worker_heartbeat_server_thread.get_address(),
            self.client_heartbeat_server_thread.get_address(),
            reply_client_ping_address, None, self.pid, self.job_id,
            self.log_server_address)

        try:
            self.job_socket.send_multipart([
                remote_constants.NORMAL_TAG,
                cloudpickle.dumps(initialized_job)
            ])
            message = self.job_socket.recv_multipart()
        except zmq.error.Again as e:
            logger.warning("[Job] Cannot connect to the worker{}. ".format(
                self.worker_address) + "Job will quit.")
            self.job_socket.close(0)
            os._exit(0)

        tag = message[0]
        assert tag == remote_constants.NORMAL_TAG
        # create the remove_job_socket
        remove_job_address = to_str(message[1])
        self.remove_job_socket = self.ctx.socket(zmq.REQ)
        self.remove_job_socket.setsockopt(
            zmq.RCVTIMEO, remote_constants.HEARTBEAT_TIMEOUT_S * 1000)
        self.remove_job_socket.connect("tcp://{}".format(remove_job_address))

    def _check_used_memory(self):
        """Check if the memory used by this job exceeds self.max_memory."""
        while True:
            if self.max_memory is not None:
                process = psutil.Process(self.pid)
                used_memory = float(process.memory_info()[0]) / (1024**2)
                if used_memory > self.max_memory + self.init_memory:
                    break
            time.sleep(remote_constants.HEARTBEAT_INTERVAL_S)

        # out of memory
        logger.error(
            "Memory used by this job exceeds {}. This job will exist.".format(
                self.max_memory))

        stop_message = "Job {} exceeds max memory usage, will stop this job.".format(
            self.job_address)
        self.client_heartbeat_server_thread.stop(
            remote_constants.HEARTBEAT_OUT_OF_MEMORY_TAG, stop_message)

    def _reply_ping(self, socket):
        """Create a socket server that reply the ping signal from client.
        This signal is used to make sure that the job is still alive.
        """
        message = socket.recv_multipart()
        max_memory = to_str(message[1])
        if max_memory != 'None':
            self.max_memory = float(max_memory)
        socket.send_multipart([remote_constants.HEARTBEAT_TAG])

        self.client_heartbeat_server_thread.start()

        memory_monitor_thread = threading.Thread(
            target=self._check_used_memory)
        memory_monitor_thread.setDaemon(True)
        memory_monitor_thread.start()

        socket.close(0)

    def wait_for_files(self, reply_socket, job_address):
        """Wait for python files from remote object.

        When a remote object receives the allocated job address, it will send
        the python files to the job. Later, the job will save these files to a
        temporary directory and add the temporary diretory to Python's working
        directory.

        Args:
            reply_socket (sockert): main socket to accept commands of remote object.
            job_address (String): address of reply_socket.

        Returns:
            A temporary directory containing the python files.
        """

        message = reply_socket.recv_multipart()
        tag = message[0]
        if tag == remote_constants.SEND_FILE_TAG:
            pyfiles = pickle.loads(message[1])
            envdir = tempfile.mkdtemp()

            for empty_subfolder in pyfiles['empty_subfolders']:
                empty_subfolder_path = os.path.join(envdir, empty_subfolder)
                if not os.path.exists(empty_subfolder_path):
                    os.makedirs(empty_subfolder_path)

            # save python files to temporary directory
            for file, code in pyfiles['python_files'].items():
                file = os.path.join(envdir, file)
                with open(file, 'wb') as code_file:
                    code_file.write(code)

            # save other files to current directory
            for file, content in pyfiles['other_files'].items():
                # create directory (i.e. ./rom_files/)
                if os.sep in file:
                    try:
                        sep = os.sep
                        recursive_dirs = os.path.join(*(file.split(sep)[:-1]))
                        recursive_dirs = os.path.join(envdir, recursive_dirs)
                        os.makedirs(recursive_dirs)
                    except OSError as e:
                        pass
                file = os.path.join(envdir, file)
                with open(file, 'wb') as f:
                    f.write(content)
            reply_socket.send_multipart([remote_constants.NORMAL_TAG])
            return envdir
        else:
            logger.error("NotImplementedError:{}, received tag:{}".format(
                job_address, tag))
            raise NotImplementedError

    def wait_for_connection(self, reply_socket):
        """Wait for connection from the remote object.

        The remote object will send its class information and initialization
        arguments to the job, these parameters are then used to create a
        local instance in the job process.

        Args:
            reply_socket (sockert): main socket to accept commands of remote object.

        Returns:
            A local instance of the remote class object.
        """

        message = reply_socket.recv_multipart()
        tag = message[0]
        obj = None

        if tag == remote_constants.INIT_OBJECT_TAG:
            try:
                file_name, class_name, end_of_file, in_sys_path = cloudpickle.loads(
                    message[1])

                # append the environment paths of the client to the current environment path.
                client_sys_path = cloudpickle.loads(message[3])
                sys.path.extend(client_sys_path)

                cls = load_remote_class(file_name, class_name, end_of_file,
                                        in_sys_path)
                args, kwargs = cloudpickle.loads(message[2])

                with redirect_output_to_file(self.logfile_path, os.devnull):
                    obj = cls(*args, **kwargs)
            except Exception as e:
                traceback_str = str(traceback.format_exc())
                error_str = str(e)
                logger.error("traceback:\n{}".format(traceback_str))
                reply_socket.send_multipart([
                    remote_constants.EXCEPTION_TAG,
                    to_byte(error_str + "\ntraceback:\n" + traceback_str)
                ])
                return None
            reply_socket.send_multipart([
                remote_constants.NORMAL_TAG,
                dumps_return(set(obj.__dict__.keys()))
            ])
        else:
            logger.error("Message from job {}".format(message))
            reply_socket.send_multipart([
                remote_constants.EXCEPTION_TAG,
                b"[job]Unkonwn tag when tried to receive the class definition"
            ])
            raise NotImplementedError

        return obj

    def run(self, job_address_sender, job_id_sender):
        """An infinite loop waiting for a new task.

        Args:
            job_address_sender(sending end of multiprocessing.Pipe): send job address of reply_socket to main process.
        """
        ctx = zmq.Context()

        # create the reply_socket
        reply_socket = ctx.socket(zmq.REP)
        job_port = reply_socket.bind_to_random_port(addr="tcp://*")
        reply_socket.linger = 0
        job_ip = get_ip_address()
        job_address = "{}:{}".format(job_ip, job_port)

        job_id = job_address.replace(':', '_') + '_' + str(int(time.time()))
        self.log_dir = os.path.expanduser('~/.parl_data/job/{}'.format(job_id))
        logger.set_dir(self.log_dir)
        self.logfile_path = os.path.join(self.log_dir, 'stdout.log')

        logger.info(
            "[Job] Job {} initialized. Reply heartbeat socket Address: {}.".
            format(job_id, job_address))

        job_address_sender.send(job_address)
        job_id_sender.send(job_id)

        # receive source code from the actor and append them to the environment variables.
        envdir = self.wait_for_files(reply_socket, job_address)
        sys.path.insert(0, envdir)
        os.chdir(envdir)

        try:
            obj = self.wait_for_connection(reply_socket)
            assert obj is not None

            self.single_task(obj, reply_socket, job_address)
        except Exception as e:
            logger.error(
                "Error occurs when running a single task. We will reset this job. \nReason:{}"
                .format(e))
            traceback_str = str(traceback.format_exc())
            logger.error("traceback:\n{}".format(traceback_str))

    def single_task(self, obj, reply_socket, job_address):
        """An infinite loop waiting for commands from the remote object.

        Each job will receive two kinds of message from the remote object:

        1. When the remote object calls a function, job will run the
           function on the local instance and return the results to the
           remote object.
        2. When the remote object is deleted, the job will quit and release
           related computation resources.

        Args:
            reply_socket (sockert): main socket to accept commands of remote object.
            job_address (String): address of reply_socket.
        """

        while True:
            message = reply_socket.recv_multipart()
            tag = message[0]
            if tag in [
                    remote_constants.CALL_TAG,
                    remote_constants.GET_ATTRIBUTE_TAG,
                    remote_constants.SET_ATTRIBUTE_TAG,
            ]:
                try:
                    if tag == remote_constants.CALL_TAG:
                        function_name = to_str(message[1])
                        data = message[2]
                        args, kwargs = loads_argument(data)

                        # Redirect stdout to stdout.log temporarily
                        with redirect_output_to_file(self.logfile_path,
                                                     os.devnull):
                            ret = getattr(obj, function_name)(*args, **kwargs)

                        ret = dumps_return(ret)
                        reply_socket.send_multipart([
                            remote_constants.NORMAL_TAG, ret,
                            dumps_return(set(obj.__dict__.keys()))
                        ])

                    elif tag == remote_constants.GET_ATTRIBUTE_TAG:
                        attribute_name = to_str(message[1])
                        with redirect_output_to_file(self.logfile_path,
                                                     os.devnull):
                            ret = getattr(obj, attribute_name)
                        ret = dumps_return(ret)
                        reply_socket.send_multipart(
                            [remote_constants.NORMAL_TAG, ret])
                    elif tag == remote_constants.SET_ATTRIBUTE_TAG:
                        attribute_name = to_str(message[1])
                        attribute_value = loads_return(message[2])
                        with redirect_output_to_file(self.logfile_path,
                                                     os.devnull):
                            setattr(obj, attribute_name, attribute_value)
                        reply_socket.send_multipart([
                            remote_constants.NORMAL_TAG,
                            dumps_return(set(obj.__dict__.keys()))
                        ])
                    else:
                        pass

                except Exception as e:
                    # reset the job

                    error_str = str(e)
                    logger.error(error_str)

                    if type(e) == AttributeError:
                        reply_socket.send_multipart([
                            remote_constants.ATTRIBUTE_EXCEPTION_TAG,
                            to_byte(error_str)
                        ])
                        raise AttributeError

                    elif type(e) == SerializeError:
                        reply_socket.send_multipart([
                            remote_constants.SERIALIZE_EXCEPTION_TAG,
                            to_byte(error_str)
                        ])
                        raise SerializeError

                    elif type(e) == DeserializeError:
                        reply_socket.send_multipart([
                            remote_constants.DESERIALIZE_EXCEPTION_TAG,
                            to_byte(error_str)
                        ])
                        raise DeserializeError

                    else:
                        traceback_str = str(traceback.format_exc())
                        logger.error("traceback:\n{}".format(traceback_str))
                        reply_socket.send_multipart([
                            remote_constants.EXCEPTION_TAG,
                            to_byte(error_str + "\ntraceback:\n" +
                                    traceback_str)
                        ])
                        break

            # receive DELETE_TAG from actor, and stop replying worker heartbeat
            elif tag == remote_constants.KILLJOB_TAG:
                reply_socket.send_multipart([remote_constants.NORMAL_TAG])
                logger.warning("An actor exits and this job {} will exit.".
                               format(job_address))
                break
            else:
                logger.error(
                    "The job receives an unknown message: {}".format(message))
                raise NotImplementedError


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--worker_address", required=True, type=str, help="worker_address")
    parser.add_argument(
        "--log_server_address",
        required=True,
        type=str,
        help="log_server_address, address of the log web server on worker")
    args = parser.parse_args()
    job = Job(args.worker_address, args.log_server_address)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys

from parl.utils import _IS_PY2


class ActorRefMonitor(object):
    def __init__(self, actor):
        """Monitor of actor reference count, which is used in the future mode.
        It checks if we should release the CPU resource of the actor or not.

        Args:
            actor(instance of ProxyWrapperNoWait)
        """
        self.actor_ref = actor

        # @TODO(zenghsh3): hardcoded
        self.actor_deleted_refcount = 5
        if _IS_PY2:
            self.actor_deleted_refcount = 6

    def is_deleted(self):
        """Check wheter the actor is deleted or out of scope
        if `self.actor_ref` is None, will always return False.

        """
        cur_refcount = sys.getrefcount(self.actor_ref)
        return cur_refcount == self.actor_deleted_refcount



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import pickle
import threading
import time
import zmq
from collections import deque, defaultdict
import parl
import sys
from parl.utils import to_str, to_byte, logger, get_ip_address
from parl.remote import remote_constants
from parl.remote.job_center import JobCenter
from parl.remote.cluster_monitor import ClusterMonitor
from parl.remote.grpc_heartbeat import HeartbeatClientThread
import cloudpickle
import time
from parl.remote.utils import get_version


class Master(object):
    """Base class for a master node, the control center for our cluster, which provides connections to workers and clients.

    There is only one master node in each cluster, and it is responsible for
    receiving jobs from the clients and allocating computation resources to
    run the jobs.

    To start a master node, we use the following xparl command line api:

    .. code-block:: python

        xparl start --port localhost:1234

    At the same time, a local worker will be started and connect to the
    master node.

    Attributes:
        job_center (JobCenter): A thread-safe data structure that stores the job address of vacant cpus.
        client_socket (zmq.Context.socket): A socket that receives submitted
                                           job from the client, and later sends
                                           job_address back to the client.
        master_ip(str): The ip address of the master node.
        cpu_num(int): The number of available CPUs in the cluster.
        worker_num(int): The number of workers connected to this cluster.
        cluster_monitor(dict): A dict to record worker status and client status.
        client_hostname(dict): A dict to store hostname for each client address.

    Args:
        port: The ip port that the master node binds to.
    """

    def __init__(self, port, monitor_port=None):
        self.ctx = zmq.Context()
        self.master_ip = get_ip_address()
        self.all_client_heartbeat_threads = []
        self.all_worker_heartbeat_threads = []
        self.monitor_url = "http://{}:{}".format(self.master_ip, monitor_port)
        logger.set_dir(
            os.path.expanduser('~/.parl_data/master/{}_{}'.format(
                self.master_ip, port)))
        self.client_socket = self.ctx.socket(zmq.REP)
        self.client_socket.bind("tcp://*:{}".format(port))
        self.client_socket.linger = 0
        self.port = port

        self.job_center = JobCenter(self.master_ip)
        self.cluster_monitor = ClusterMonitor()
        self.master_is_alive = True
        self.client_hostname = defaultdict(int)

    def _get_status(self):
        return self.cluster_monitor.get_status()

    def _print_workers(self):
        """Display `worker_pool` infomation."""
        logger.info(
            "Master connects to {} workers and have {} vacant CPUs.\n".format(
                self.worker_num, self.cpu_num))

    @property
    def cpu_num(self):
        return self.job_center.cpu_num

    @property
    def worker_num(self):
        return self.job_center.worker_num

    def _receive_message(self):
        """Master node will receive various types of message: (1) worker
        connection; (2) worker update; (3) client connection; (4) job
        submittion; (5) reset job.
        """
        message = self.client_socket.recv_multipart()
        tag = message[0]

        # a new worker connects to the master
        if tag == remote_constants.WORKER_CONNECT_TAG:
            self.client_socket.send_multipart([remote_constants.NORMAL_TAG])

        elif tag == remote_constants.MONITOR_TAG:
            status = self._get_status()
            self.client_socket.send_multipart(
                [remote_constants.NORMAL_TAG, status])

        # `xparl status` command line API
        elif tag == remote_constants.STATUS_TAG:
            status_info = self.cluster_monitor.get_status_info()
            self.client_socket.send_multipart(
                [remote_constants.NORMAL_TAG,
                 to_byte(status_info)])

        elif tag == remote_constants.WORKER_INITIALIZED_TAG:
            initialized_worker = cloudpickle.loads(message[1])
            worker_address = initialized_worker.worker_address
            self.job_center.add_worker(initialized_worker)
            hostname = self.job_center.get_hostname(worker_address)
            self.cluster_monitor.add_worker_status(worker_address, hostname)
            logger.info("A new worker {} is added, ".format(worker_address) +
                        "the cluster has {} CPUs.\n".format(self.cpu_num))

            def heartbeat_exit_callback_func(worker_address):
                self.job_center.drop_worker(worker_address)
                self.cluster_monitor.drop_worker_status(worker_address)
                logger.warning("\n[Master] Cannot connect to the worker " +
                               "{}. ".format(worker_address) +
                               "Worker_pool will drop this worker.")
                self._print_workers()
                logger.warning("Exit worker monitor from master.")

            # a thread for sending heartbeat signals to the client
            thread = HeartbeatClientThread(
                worker_address,
                heartbeat_exit_callback_func=heartbeat_exit_callback_func,
                exit_func_args=(worker_address, ))
            self.all_worker_heartbeat_threads.append(thread)
            thread.setDaemon(True)
            thread.start()

            self.client_socket.send_multipart([remote_constants.NORMAL_TAG])

        # a client connects to the master
        elif tag == remote_constants.CLIENT_CONNECT_TAG:
            # `client_heartbeat_address` is the
            #      `reply_master_heartbeat_address` of the client

            client_heartbeat_address = to_str(message[1])
            client_hostname = to_str(message[2])
            client_id = to_str(message[3])
            self.client_hostname[client_heartbeat_address] = client_hostname
            logger.info(
                "Client {} is connected.".format(client_heartbeat_address))

            def heartbeat_exit_callback_func(client_heartbeat_address):
                self.cluster_monitor.drop_client_status(
                    client_heartbeat_address)
                logger.warning("[Master] cannot connect to the client " +
                               "{}. ".format(client_heartbeat_address) +
                               "Please check if it is still alive.")
                logger.info(
                    "Master connects to {} workers and have {} vacant CPUs.\n".
                    format(self.worker_num, self.cpu_num))

            # a thread for sending heartbeat signals to the client
            thread = HeartbeatClientThread(
                client_heartbeat_address,
                heartbeat_exit_callback_func=heartbeat_exit_callback_func,
                exit_func_args=(client_heartbeat_address, ))
            self.all_client_heartbeat_threads.append(thread)
            thread.setDaemon(True)
            thread.start()

            log_monitor_address = "{}/logs?client_id={}".format(
                self.monitor_url, client_id)
            self.client_socket.send_multipart(
                [remote_constants.NORMAL_TAG,
                 to_byte(log_monitor_address)])

        elif tag == remote_constants.CHECK_VERSION_TAG:
            self.client_socket.send_multipart([
                remote_constants.NORMAL_TAG,
                to_byte(parl.__version__),
                to_byte(str(sys.version_info.major)),
                to_byte(str(sys.version_info.minor)),
                to_byte(str(get_version('pyarrow')))
            ])

        # a client submits a job to the master
        elif tag == remote_constants.CLIENT_SUBMIT_TAG:
            # check available CPU resources
            if self.cpu_num:
                logger.info("Submitting job...")
                job = self.job_center.request_job()
                self.client_socket.send_multipart([
                    remote_constants.NORMAL_TAG,
                    to_byte(job.job_address),
                    to_byte(job.client_heartbeat_address),
                    to_byte(job.ping_heartbeat_address),
                ])
                client_id = to_str(message[2])
                job_info = {job.job_id: job.log_server_address}
                self.cluster_monitor.add_client_job(client_id, job_info)
                self._print_workers()
            else:
                self.client_socket.send_multipart([remote_constants.CPU_TAG])

        # a worker updates
        elif tag == remote_constants.NEW_JOB_TAG:
            initialized_job = cloudpickle.loads(message[1])
            last_job_address = to_str(message[2])

            self.client_socket.send_multipart([remote_constants.NORMAL_TAG])
            self.job_center.update_job(last_job_address, initialized_job,
                                       initialized_job.worker_address)
            logger.info("A worker updated. cpu_num:{}".format(self.cpu_num))

            self._print_workers()

        # client update status periodically
        elif tag == remote_constants.CLIENT_STATUS_UPDATE_TAG:
            client_heartbeat_address = to_str(message[1])
            client_status = cloudpickle.loads(message[2])

            client_status['client_hostname'] = self.client_hostname[
                client_heartbeat_address]
            self.cluster_monitor.update_client_status(client_heartbeat_address,
                                                      client_status)
            self.client_socket.send_multipart([remote_constants.NORMAL_TAG])

        # worker update status periodically
        elif tag == remote_constants.WORKER_STATUS_UPDATE_TAG:
            worker_address = to_str(message[1])
            worker_status = cloudpickle.loads(message[2])

            vacant_cpus = self.job_center.get_vacant_cpu(worker_address)
            total_cpus = self.job_center.get_total_cpu(worker_address)
            self.cluster_monitor.update_worker_status(
                worker_status, worker_address, vacant_cpus, total_cpus)

            self.client_socket.send_multipart([remote_constants.NORMAL_TAG])

        # check before start a worker
        elif tag == remote_constants.NORMAL_TAG:
            self.client_socket.send_multipart([remote_constants.NORMAL_TAG])

        else:
            raise NotImplementedError()

    def exit(self):
        """ Close the master.
        """
        self.master_is_alive = False

        for thread in self.all_client_heartbeat_threads:
            if thread.is_alive():
                thread.exit()

        for thread in self.all_worker_heartbeat_threads:
            if thread.is_alive():
                thread.exit()

    def run(self):
        """An infinite loop waiting for messages from the workers and
        clients.

        Master node will receive four types of messages:

        1. A new worker connects to the master node.
        2. A connected worker sending new job address after it kills an old
           job.
        3. A new client connects to the master node.
        4. A connected client submits a job after a remote object is created.
        """
        self.client_socket.linger = 0
        self.client_socket.setsockopt(
            zmq.RCVTIMEO, remote_constants.HEARTBEAT_RCVTIMEO_S * 1000)

        while self.master_is_alive:
            try:
                self._receive_message()
                pass
            except zmq.error.Again as e:
                #detect whether `self.master_is_alive` is True periodically
                pass

        logger.warning("[Master] Exit master.")



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import linecache
import os
import zmq

from flask import Flask, current_app, jsonify, make_response, request, send_file
from flask_cors import CORS
from parl.remote import remote_constants
from parl.utils import to_byte, logger
from parl.remote.grpc_heartbeat import HeartbeatServerThread
from parl.remote.zmq_utils import create_client_socket

app = Flask(__name__)
CORS(app)


@app.route(
    "/get-log", methods=[
        'GET',
    ])
def get_log():
    '''
    args: 
        job_id: id of the remote job
    response:
        log: newest `LINE_NUM` lines of the log file
    '''
    try:
        job_id = request.args['job_id']
    except:
        return make_response(
            jsonify(message="No job_id provided, please check your request."),
            400)

    log_dir = current_app.config.get('LOG_DIR')
    log_dir = os.path.expanduser(log_dir)
    log_file_path = os.path.join(log_dir, job_id, 'stdout.log')
    if not os.path.isfile(log_file_path):
        return make_response(
            jsonify(message="Log not exsits, please check your job_id"), 400)
    else:
        line_num = current_app.config.get('LINE_NUM')
        linecache.checkcache(log_file_path)
        log_content = ''.join(linecache.getlines(log_file_path)[-line_num:])
        return make_response(
            jsonify(message="Log exsits, content in log", log=log_content),
            200)


@app.route(
    '/download-log', methods=[
        'GET',
    ])
def download_log():
    '''
    args:
        job_id: the id of the remote job
    response:
        log: log file
    '''
    try:
        job_id = request.args['job_id']
    except:
        return make_response(
            jsonify(message="No job_id provided, please check your request."),
            400)
    log_dir = current_app.config.get('LOG_DIR')
    log_dir = os.path.expanduser(log_dir)
    log_file_path = os.path.join(log_dir, job_id, 'stdout.log')
    if not os.path.isfile(log_file_path):
        return make_response(
            jsonify(message="Log not exsits, please check your job_id"), 400)
    else:
        return send_file(log_file_path, as_attachment=True)


def send_heartbeat_addr_to_worker(worker_addr, heartbeat_server_addr):
    ctx = zmq.Context()
    socket = create_client_socket(ctx, worker_addr, heartbeat_timeout=True)

    try:
        socket.send_multipart([
            remote_constants.NORMAL_TAG,
            to_byte(heartbeat_server_addr),
        ])
        message = socket.recv_multipart()
    except zmq.error.Again as e:
        err_str = "Can not connect to the worker please " \
                  "check if the worker is running."
        logger.warning(err_str)
        raise Exception(err_str)


if __name__ == "__main__":
    import logging
    log = logging.getLogger('werkzeug')
    log.disabled = True

    parser = argparse.ArgumentParser()
    parser.add_argument('--port', required=True, type=int)
    parser.add_argument('--log_dir', required=True, type=str)
    parser.add_argument('--line_num', required=True, type=int)
    parser.add_argument('--worker_address', required=True, type=str)
    args = parser.parse_args()

    app.config.from_mapping(
        LOG_DIR=args.log_dir,
        LINE_NUM=args.line_num,
    )

    def heartbeat_exit_callback_func():
        logger.warning(
            "[log_server] lost connnect with the worker. Please check if it is still alive."
        )
        os._exit(1)

    heartbeat_server_thread = HeartbeatServerThread(
        heartbeat_exit_callback_func=heartbeat_exit_callback_func)
    heartbeat_server_thread.setDaemon(True)
    heartbeat_server_thread.start()

    send_heartbeat_addr_to_worker(args.worker_address,
                                  heartbeat_server_thread.get_address())

    app.run(host="0.0.0.0", port=args.port)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import torch.nn as nn

from parl.core.model_base import ModelBase
from parl.utils import machine_info

__all__ = ['Model']


class Model(nn.Module, ModelBase):
    """
    | `alias`: ``parl.Model``
    | `alias`: ``parl.core.torch.agent.Model``

    | ``Model`` is a base class of PARL for the neural network. A ``Model`` is
    usually a policy or Q-value function, which predicts an action or an
    estimate according to the environmental observation.

    | To use the ``PyTorch`` backend model, user needs to call ``super(Model,
    self).__init__()`` at the beginning of ``__init__`` function.

    | ``Model`` supports duplicating a ``Model`` instance in a pythonic way:

    | ``copied_model = copy.deepcopy(model)``

    Example:

    .. code-block:: python

        import parl
        import torch.nn as nn

        class Policy(parl.Model):
            def __init__(self):
                super(Policy, self).__init__()
                self.fc = nn.Linear(in_features=100, out_features=32)

            def policy(self, obs):
                out = self.fc(obs)
                return out

        policy = Policy()
        copied_policy = copy.deepcopy(model)

    Attributes:
        model_id(str): each model instance has its unique model_id.

    Public Functions:
        - ``sync_weights_to``: synchronize parameters of the current model to
        another model.
        - ``get_weights``: return a dict containing all the parameters of the
        current model.
        - ``set_weights``: copy parameters from ``set_weights()`` to the model.
        - ``forward``: define the computations of a neural network. **Should**
        be overridden by all subclasses.

    """

    def __init___(self):
        super(Model, self).__init__()

    def sync_weights_to(self, target_model, decay=0.0):
        """Synchronize parameters of current model to another model.

        target_model_weights = decay * target_model_weights + (1 - decay) *
        current_model_weights

        Args:
            target_model (`parl.Model`): an instance of ``Model`` that has the
            same neural network architecture as the current model.
            decay (float):  the rate of decline in copying parameters. 0 if no
            parameters decay when synchronizing the parameters.

        Example:

        .. code-block:: python

            import copy
            # create a model that has the same neural network structures.
            target_model = copy.deepcopy(model)

            # after initializing the parameters ...
            model.sync_weights_to(target_model)

        Note:
            Before calling ``sync_weights_to``, parameters of the model must
            have been initialized.
        """

        assert not target_model is self, "cannot copy between identical model"
        assert isinstance(target_model, Model)
        assert self.__class__.__name__ == target_model.__class__.__name__, \
            "must be the same class for params syncing!"
        assert (decay >= 0 and decay <= 1)

        target_vars = dict(target_model.named_parameters())
        for name, var in self.named_parameters():
            target_vars[name].data.copy_(decay * target_vars[name].data +
                                         (1 - decay) * var.data)

    def get_weights(self):
        """Returns a Python dict containing parameters of current model.

        Returns: a Python dict containing the parameters of current model.
        """
        weights = self.state_dict()
        for key in weights.keys():
            weights[key] = weights[key].cpu().numpy()
        return weights

    def set_weights(self, weights):
        """Copy parameters from ``set_weights()`` to the model.
        
        Args:
            weights (dict): a Python dict containing the parameters.
        """
        new_weights = dict()
        for key in weights.keys():
            new_weights[key] = torch.from_numpy(weights[key])
        self.load_state_dict(new_weights)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import threading
from parl.utils import logger
import signal
import os


class WorkerStatus(object):
    """Maintain worker's information in a worker node. 

    Attributes:
        cpu_num(int): The number of CPUs to be used in this worker.
        jobs(set): A set that records job addresses provided to the master.
        worker_address(str): Address of the worker.
    """

    def __init__(self, worker_address, initialized_jobs, cpu_num):
        self.worker_address = worker_address
        self.jobs = dict()
        for job in initialized_jobs:
            self.jobs[job.job_address] = job
        self._lock = threading.Lock()
        self.cpu_num = cpu_num

    def remove_job(self, killed_job):
        """Rmove a job from internal job pool.

        Args:
            killed_job(str): Job address to be removed.

        Returns: True if removing the job succeeds.
        """
        ret = False
        self._lock.acquire()
        if killed_job in self.jobs:
            pid = self.jobs[killed_job].pid
            self.jobs.pop(killed_job)
            ret = True
            try:
                os.kill(pid, signal.SIGTERM)
            except OSError:
                logger.warning("job:{} has been killed before".format(pid))
            logger.info("[Worker] kills a job:{}".format(killed_job))
        self._lock.release()
        return ret

    def clear(self):
        """Remove all the jobs"""
        self._lock.acquire()
        for job in self.jobs.values():
            try:
                os.kill(job.pid, signal.SIGTERM)
            except OSError:
                logger.warning("job:{} has been killed before".format(job.pid))
            logger.info("[Worker] kills a job:{}".format(job.pid))
        self.jobs = dict()
        self._lock.release()

    def add_job(self, new_job):
        """Add a new job to interal job pool.
        
        Args:
            new_job(InitializedJob): Initialized job to be added to the self.jobs.
        """
        self._lock.acquire()
        self.jobs[new_job.job_address] = new_job
        assert len(self.jobs) <= self.cpu_num
        self._lock.release()



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import pickle
import random
import time
import zmq
import threading

from flask import Flask, render_template, jsonify, request

app = Flask(__name__)


@app.route('/')
@app.route('/workers')
def worker():
    return render_template('workers.html')


@app.route('/clients')
def clients():
    return render_template('clients.html')


class ClusterMonitor(object):
    """A monitor which requests the cluster status every 10 seconds.
    """

    def __init__(self, master_address):
        ctx = zmq.Context()
        self.socket = ctx.socket(zmq.REQ)
        self.socket.setsockopt(zmq.RCVTIMEO, 30000)
        self.socket.connect('tcp://{}'.format(master_address))
        self.data = None

        thread = threading.Thread(target=self.run)
        thread.setDaemon(True)
        thread.start()

    def run(self):
        master_is_alive = True
        while master_is_alive:
            try:
                self.socket.send_multipart([b'[MONITOR]'])
                msg = self.socket.recv_multipart()

                status = pickle.loads(msg[1])
                data = {'workers': [], 'clients': []}
                total_vacant_cpus = 0
                total_used_cpus = 0

                master_idx = None
                for idx, worker in enumerate(status['workers'].values()):
                    worker['load_time'] = list(worker['load_time'])
                    worker['load_value'] = list(worker['load_value'])
                    if worker['hostname'] == 'Master':
                        master_idx = idx
                    data['workers'].append(worker)
                    total_used_cpus += worker[
                        'used_cpus'] if 'used_cpus' in worker else 0
                    total_vacant_cpus += worker[
                        'vacant_cpus'] if 'vacant_cpus' in worker else 0

                if master_idx != 0 and master_idx is not None:
                    master_worker = data['workers'].pop(master_idx)
                    data['workers'] = [master_worker] + data['workers']

                data['total_vacant_cpus'] = total_vacant_cpus
                data['total_cpus'] = total_used_cpus + total_vacant_cpus
                data['clients'] = list(status['clients'].values())
                data['client_jobs'] = status['client_jobs']
                self.data = data
                time.sleep(10)

            except zmq.error.Again as e:
                master_is_alive = False
                self.socket.close(0)

    def get_data(self):
        assert self.data is not None
        return self.data


@app.route('/cluster')
def cluster():
    data = CLUSTER_MONITOR.get_data()
    return jsonify(data)


@app.route(
    '/logs', methods=[
        'GET',
    ])
def logs():
    client_id = request.args.get('client_id')
    return render_template('jobs.html', client_id=client_id)


@app.route(
    '/get-jobs', methods=[
        'GET',
    ])
def get_jobs():
    client_id = request.args.get('client_id')
    jobs = CLUSTER_MONITOR.get_data()['client_jobs'].get(client_id)
    data = []
    if jobs:
        for idx, job_id in enumerate(jobs):
            monitor_url = jobs[job_id]
            data.append({
                "id":
                idx,
                "job_id":
                job_id,
                "log_url":
                "http://{}/get-log?job_id={}".format(monitor_url, job_id),
                "download_url":
                "http://{}/download-log?job_id={}".format(monitor_url, job_id),
            })
    return jsonify(data)


if __name__ == "__main__":
    import logging
    log = logging.getLogger('werkzeug')
    log.disabled = True

    parser = argparse.ArgumentParser()
    parser.add_argument('--monitor_port', default=1234, type=int)
    parser.add_argument('--address', default='localhost:8010', type=str)
    args = parser.parse_args()

    CLUSTER_MONITOR = ClusterMonitor(args.address)
    app.run(host="0.0.0.0", port=args.monitor_port)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import locale
import multiprocessing
import os
import random
import re
import requests
import subprocess
import sys
import time
import threading
import tempfile
import warnings
import zmq
from multiprocessing import Process
from parl.utils import (_IS_WINDOWS, get_free_tcp_port, get_ip_address,
                        get_port_from_range, is_port_available, kill_process,
                        to_str)
from parl.remote.remote_constants import STATUS_TAG

# A flag to mark if parl is started from a command line
os.environ['XPARL'] = 'True'

# Solve `Click will abort further execution because Python 3 was configured
# to use ASCII as encoding for the environment` error.

if not _IS_WINDOWS:
    try:
        locale.setlocale(locale.LC_ALL, "en_US.UTF-8")
    except:
        pass

#TODO: this line will cause error in python2/macOS
if sys.version_info.major == 3:
    warnings.simplefilter("ignore", ResourceWarning)


def is_master_started(address):
    ctx = zmq.Context()
    socket = ctx.socket(zmq.REQ)
    socket.linger = 0
    socket.setsockopt(zmq.RCVTIMEO, 500)
    socket.connect("tcp://{}".format(address))
    socket.send_multipart([b'[NORMAL]'])
    try:
        _ = socket.recv_multipart()
        socket.close(0)
        return True
    except zmq.error.Again as e:
        socket.close(0)
        return False


def parse_port_range(log_server_port_range):
    try:
        re.match(r'\d*[-]\d*', log_server_port_range).span()
    except:
        raise Exception(
            "The input log_server_port_range should be `start-end` format.")
    start, end = map(int, log_server_port_range.split('-'))
    if start > end:
        raise Exception(
            "Start port number must be smaller than the end port number.")

    return start, end


def check_log_server_started(ip_address, port):
    started = False
    for _ in range(3):
        try:
            r = requests.get("http://{}:{}/get-log".format(ip_address, port))
            if r.status_code == 400:
                started = True
                break
        except:
            pass
        time.sleep(3)
        click.echo("Checking status of log_server...")

    if started:
        click.echo("# Start the log server sucessfully.")
    else:
        click.echo("# Fail to start the log server.")


@click.group()
def cli():
    pass


@click.command("start", short_help="Start a master node.")
@click.option("--port", help="The port to bind to.", type=str, required=True)
@click.option(
    "--debug",
    help="Start parl in the debugging mode to print all running log.",
    is_flag=True)
@click.option(
    "--cpu_num",
    type=int,
    help="Set number of cpu manually. If not set, it will use all "
    "cpus of this machine.")
@click.option(
    "--monitor_port", help="The port to start a cluster monitor.", type=str)
@click.option(
    "--log_server_port_range",
    help='''
    Port range (start-end) of the log server on the worker. Default: 8000-9000. 
    The worker will pick a random avaliable port in [start, end] for the log server.
    ''',
    default="8000-9000",
    type=str)
def start_master(port, cpu_num, monitor_port, debug, log_server_port_range):
    if cpu_num is not None:
        assert cpu_num >= 0, "cpu_num should be greater than or equal to 0."
    if debug:
        os.environ['DEBUG'] = 'True'

    if not is_port_available(port):
        raise Exception(
            "The master address localhost:{} is already in use.".format(port))

    if monitor_port and not is_port_available(monitor_port):
        raise Exception(
            "The input monitor port localhost:{} is already in use.".format(
                monitor_port))

    cpu_num = int(
        cpu_num) if cpu_num is not None else multiprocessing.cpu_count()
    start_file = __file__.replace('scripts.pyc', 'start.py')
    start_file = start_file.replace('scripts.py', 'start.py')
    monitor_file = __file__.replace('scripts.pyc', 'monitor.py')
    monitor_file = monitor_file.replace('scripts.py', 'monitor.py')

    monitor_port = monitor_port if monitor_port else get_free_tcp_port()
    start, end = parse_port_range(log_server_port_range)
    log_server_port = get_port_from_range(start, end)
    while log_server_port == monitor_port or log_server_port == port:
        log_server_port = get_port_from_range(start, end)

    master_command = [
        sys.executable,
        start_file,
        "--name",
        "master",
        "--port",
        port,
        "--monitor_port",
        monitor_port,
    ]
    worker_command = [
        sys.executable, start_file, "--name", "worker", "--address",
        "localhost:" + str(port), "--cpu_num",
        str(cpu_num), '--log_server_port',
        str(log_server_port)
    ]
    monitor_command = [
        sys.executable, monitor_file, "--monitor_port",
        str(monitor_port), "--address", "localhost:" + str(port)
    ]

    FNULL = open(os.devnull, 'w')

    # Redirect the output to DEVNULL to solve the warning log.
    _ = subprocess.Popen(master_command, stdout=FNULL, close_fds=True)

    if cpu_num > 0:
        # Sleep 1s for master ready
        time.sleep(1)
        _ = subprocess.Popen(worker_command, stdout=FNULL, close_fds=True)

    if _IS_WINDOWS:
        # TODO(@zenghsh3) redirecting stdout of monitor subprocess to FNULL will cause occasional failure
        tmp_file = tempfile.TemporaryFile()
        _ = subprocess.Popen(monitor_command, stdout=tmp_file, close_fds=True)
        tmp_file.close()
    else:
        _ = subprocess.Popen(monitor_command, stdout=FNULL, close_fds=True)
    FNULL.close()

    if cpu_num > 0:
        monitor_info = """
            # The Parl cluster is started at localhost:{}.

            # A local worker with {} CPUs is connected to the cluster.    

            # Starting the cluster monitor...""".format(
            port,
            cpu_num,
        )
    else:
        monitor_info = """
            # The Parl cluster is started at localhost:{}.

            # Starting the cluster monitor...""".format(port)
    click.echo(monitor_info)

    # check if monitor is started
    monitor_is_started = False
    if _IS_WINDOWS:
        cmd = r'''wmic process where "commandline like '%remote\\monitor.py --monitor_port {} --address localhost:{}%'" get commandline /format:list | findstr /V wmic | findstr CommandLine='''.format(
            monitor_port, port)
    else:
        cmd = r'ps -ef | grep -v grep | grep remote/monitor.py\ --monitor_port\ {}\ --address\ localhost:{}'.format(
            monitor_port, port)
    for i in range(3):
        check_monitor_is_started = os.popen(cmd).read()
        if len(check_monitor_is_started) > 0:
            monitor_is_started = True
            break
        time.sleep(3)

    master_ip = get_ip_address()
    if monitor_is_started:
        start_info = """
        ## If you want to check cluster status, please view:

            http://{}:{}

        or call:

            xparl status""".format(master_ip, monitor_port)
    else:
        start_info = "# Fail to start the cluster monitor."

    monitor_info = """
        {}

        ## If you want to add more CPU resources, please call:

            xparl connect --address {}:{}

        ## If you want to shutdown the cluster, please call:

            xparl stop        
        """.format(start_info, master_ip, port)
    click.echo(monitor_info)

    if cpu_num > 0:
        check_log_server_started(master_ip, log_server_port)


@click.command("connect", short_help="Start a worker node.")
@click.option(
    "--address", help="IP address of the master node.", required=True)
@click.option(
    "--cpu_num",
    type=int,
    help="Set number of cpu manually. If not set, it will use all "
    "cpus of this machine.")
@click.option(
    "--log_server_port_range",
    help='''
    Port range (start-end) of the log server on the worker. Default: 8000-9000. 
    The worker will pick a random avaliable port in [start, end] for the log server.
    ''',
    default="8000-9000",
    type=str)
def start_worker(address, cpu_num, log_server_port_range):
    if cpu_num is not None:
        assert cpu_num > 0, "cpu_num should be greater than 0."

    start, end = parse_port_range(log_server_port_range)
    log_server_port = get_port_from_range(start, end)

    if not is_master_started(address):
        raise Exception("Worker can not connect to the master node, " +
                        "please check if the input address {} ".format(
                            address) + "is correct.")
    cpu_num = str(cpu_num) if cpu_num else ''
    start_file = __file__.replace('scripts.pyc', 'start.py')
    start_file = start_file.replace('scripts.py', 'start.py')

    command = [
        sys.executable, start_file, "--name", "worker", "--address", address,
        "--cpu_num",
        str(cpu_num), "--log_server_port",
        str(log_server_port)
    ]
    FNULL = open(os.devnull, 'w')
    p = subprocess.Popen(command, stdout=FNULL, close_fds=True)
    FNULL.close()

    check_log_server_started(get_ip_address(), log_server_port)


@click.command("stop", help="Exit the cluster.")
def stop():
    kill_process('remote/start.py')
    kill_process('remote/job.py')
    kill_process('remote/monitor.py')
    kill_process('remote/log_server.py')


@click.command("status")
def status():
    if _IS_WINDOWS:
        cmd = r'''wmic process where "commandline like '%remote\\start.py --name worker --address%'" get commandline /format:list | findstr /V wmic | findstr CommandLine='''
    else:
        cmd = r'ps -ef | grep remote/start.py\ --name\ worker\ --address'

    content = os.popen(cmd).read().strip()
    pattern = re.compile('--address (.*?) --cpu')
    clusters = set(pattern.findall(content))
    if len(clusters) == 0:
        click.echo('No active cluster is found.')
    else:
        ctx = zmq.Context()
        status = []
        for cluster in clusters:
            if _IS_WINDOWS:
                cmd = r'''wmic process where "commandline like '%address {}%'" get commandline /format:list | findstr /V wmic | findstr CommandLine='''.format(
                    cluster)
            else:
                cmd = r'ps -ef | grep address\ {}'.format(cluster)
            content = os.popen(cmd).read()
            pattern = re.compile('--monitor_port (.*?)\n', re.S)
            monitors = pattern.findall(content)

            if len(monitors):
                monitor_port, _, master_address = monitors[0].split(' ')
                monitor_address = "{}:{}".format(get_ip_address(),
                                                 monitor_port)
                socket = ctx.socket(zmq.REQ)
                socket.setsockopt(zmq.RCVTIMEO, 10000)
                socket.connect('tcp://{}'.format(master_address))
                try:
                    socket.send_multipart([STATUS_TAG])
                    monitor_info = to_str(socket.recv_multipart()[1])
                except zmq.error.Again as e:
                    click.echo(
                        'Can not connect to cluster {}, please try later.'.
                        format(master_address))
                    socket.close(0)
                    continue
                msg = """
            # Cluster {} {}

            # If you want to check cluster status, please view: http://{}
            """.format(master_address, monitor_info, monitor_address)
                status.append(msg)
                socket.close(0)
            else:
                msg = """
            # Cluster {} fails to start the cluster monitor.
                """.format(cluster)
                status.append(msg)

        for monitor_status in status:
            click.echo(monitor_status)


cli.add_command(start_worker)
cli.add_command(start_master)
cli.add_command(stop)
cli.add_command(status)


def main():
    return cli()


if __name__ == "__main__":
    main()



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import threading
from parl.utils.utils import _IS_PY2

__all__ = ['CSVLogger']


class CSVLogger(object):
    def __init__(self, output_file):
        """CSV Logger which can write dict result to csv file.

        Args:
            output_file(str): filename of the csv file.

        Examples:
            ```python
            from parl.utils import CSVLogger

            csv_logger = CSVLogger("result.csv")
            csv_logger.log_dict({"loss": 1, "reward": 2})
            csv_logger.log_dict({"loss": 3, "reward": 4})
            ```

            The content of the `result.csv`:
            loss,reward
            1,2
            3,4
        """

        # reference: https://stackoverflow.com/questions/1170214/python-2-csv-writer-produces-wrong-line-terminator-on-windows/1170297#1170297
        if _IS_PY2:
            self.output_file = open(output_file, "wb")
        else:
            self.output_file = open(output_file, "w", newline="")

        self.csv_writer = None
        self.lock = threading.Lock()
        self.keys_set = None

    def log_dict(self, result):
        """Ouput result to the csv file.

        Will create the header of the csv file automatically when the function is called for the first time.
        Ususally, the keys of the result should be the same every time you call the function.

        Args:
            result(dict)
        """
        with self.lock:
            assert isinstance(result, dict), "the input should be a dict."
            if self.csv_writer is None:
                self.csv_writer = csv.DictWriter(self.output_file,
                                                 result.keys())
                self.csv_writer.writeheader()
                self.keys_set = set(result.keys())

            assert set(
                result.keys()
            ) == self.keys_set, "The keys of the dict should be the same as before."

            self.csv_writer.writerow({
                k: v
                for k, v in result.items() if k in self.csv_writer.fieldnames
            })

    def flush(self):
        with self.lock:
            self.output_file.flush()

    def close(self):
        with self.lock:
            if not self.output_file.closed:
                self.output_file.close()

    def __del__(self):
        with self.lock:
            if not self.output_file.closed:
                self.output_file.close()



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from parl.remote.grpc_heartbeat import heartbeat_pb2 as heartbeat__pb2


class GrpcHeartbeatStub(object):
    """The hearbeat service definition.
    """

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Send = channel.unary_unary(
            '/GrpcHeartbeat/Send',
            request_serializer=heartbeat__pb2.Request.SerializeToString,
            response_deserializer=heartbeat__pb2.Reply.FromString,
        )


class GrpcHeartbeatServicer(object):
    """The hearbeat service definition.
    """

    def Send(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GrpcHeartbeatServicer_to_server(servicer, server):
    rpc_method_handlers = {
        'Send':
        grpc.unary_unary_rpc_method_handler(
            servicer.Send,
            request_deserializer=heartbeat__pb2.Request.FromString,
            response_serializer=heartbeat__pb2.Reply.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        'GrpcHeartbeat', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler, ))


# This class is part of an EXPERIMENTAL API.
class GrpcHeartbeat(object):
    """The hearbeat service definition.
    """

    @staticmethod
    def Send(request,
             target,
             options=(),
             channel_credentials=None,
             call_credentials=None,
             insecure=False,
             compression=None,
             wait_for_ready=None,
             timeout=None,
             metadata=None):
        return grpc.experimental.unary_unary(
            request, target, '/GrpcHeartbeat/Send',
            heartbeat__pb2.Request.SerializeToString,
            heartbeat__pb2.Reply.FromString, options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout,
            metadata)



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class InitializedJob(object):
    def __init__(self,
                 job_address,
                 worker_heartbeat_address,
                 client_heartbeat_address,
                 ping_heartbeat_address,
                 worker_address,
                 pid,
                 job_id=None,
                 log_server_address=None):
        """
    Args:
      job_address(str): Job address to which the new task connect.
      worker_heartbeat_address(str): Optional. The address to which the worker sends heartbeat signals.
      client_heartbeat_address(str): Address to which the client sends heartbeat signals.
      ping_heartbeat_address(str): the server address to which the client sends ping signals.
                                    The signal is used to check if the job is alive.
      worker_address(str): Worker's server address that receive command from the master.
      pid(int): Optional. Process id of the job.
      is_alive(True): Optional. This flag is used in worker to make sure that only alive jobs can be added into the worker_status.
    """
        self.job_address = job_address
        self.worker_heartbeat_address = worker_heartbeat_address
        self.client_heartbeat_address = client_heartbeat_address
        self.ping_heartbeat_address = ping_heartbeat_address
        self.worker_address = worker_address
        self.pid = pid
        self.is_alive = True
        self.job_id = job_id
        self.log_server_address = log_server_address


class InitializedWorker(object):
    def __init__(self, master_heartbeat_address, initialized_jobs, cpu_num,
                 hostname):
        """
    Args:
      worker_address(str): Worker server address that receives commands from the master.
      master_heartbeat_address(str): Address to which the worker send heartbeat signals to.
      initialized_jobs(list): A list of ``InitializedJob`` containing the information for initialized jobs.
      cpu_num(int): The number of CPUs used in this worker.
    """
        self.worker_address = master_heartbeat_address
        self.initialized_jobs = initialized_jobs
        self.cpu_num = cpu_num
        self.hostname = hostname



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from parl.utils import logger

__all__ = ['ReplayMemory']


class ReplayMemory(object):
    def __init__(self, max_size, obs_dim, act_dim):
        self.max_size = int(max_size)
        self.obs_dim = obs_dim
        self.act_dim = act_dim

        self.obs = np.zeros((max_size, obs_dim), dtype='float32')
        if act_dim == 0:  # Discrete control environment
            self.action = np.zeros((max_size, ), dtype='int32')
        else:  # Continuous control environment
            self.action = np.zeros((max_size, act_dim), dtype='float32')
        self.reward = np.zeros((max_size, ), dtype='float32')
        self.terminal = np.zeros((max_size, ), dtype='bool')
        self.next_obs = np.zeros((max_size, obs_dim), dtype='float32')

        self._curr_size = 0
        self._curr_pos = 0

    def sample_batch(self, batch_size):
        batch_idx = np.random.randint(self._curr_size, size=batch_size)

        obs = self.obs[batch_idx]
        reward = self.reward[batch_idx]
        action = self.action[batch_idx]
        next_obs = self.next_obs[batch_idx]
        terminal = self.terminal[batch_idx]
        return obs, action, reward, next_obs, terminal

    def make_index(self, batch_size):
        batch_idx = np.random.randint(self._curr_size, size=batch_size)
        return batch_idx

    def sample_batch_by_index(self, batch_idx):
        obs = self.obs[batch_idx]
        reward = self.reward[batch_idx]
        action = self.action[batch_idx]
        next_obs = self.next_obs[batch_idx]
        terminal = self.terminal[batch_idx]
        return obs, action, reward, next_obs, terminal

    def append(self, obs, act, reward, next_obs, terminal):
        if self._curr_size < self.max_size:
            self._curr_size += 1
        self.obs[self._curr_pos] = obs
        self.action[self._curr_pos] = act
        self.reward[self._curr_pos] = reward
        self.next_obs[self._curr_pos] = next_obs
        self.terminal[self._curr_pos] = terminal
        self._curr_pos = (self._curr_pos + 1) % self.max_size

    def size(self):
        return self._curr_size

    def __len__(self):
        return self._curr_size

    def save(self, pathname):
        other = np.array([self._curr_size, self._curr_pos], dtype=np.int32)
        np.savez(
            pathname,
            obs=self.obs,
            action=self.action,
            reward=self.reward,
            terminal=self.terminal,
            next_obs=self.next_obs,
            other=other)

    def load(self, pathname):
        data = np.load(pathname)
        other = data['other']
        if int(other[0]) > self.max_size:
            logger.warn('loading from a bigger size rpm!')
        self._curr_size = min(int(other[0]), self.max_size)
        self._curr_pos = min(int(other[1]), self.max_size - 1)

        self.obs[:self._curr_size] = data['obs'][:self._curr_size]
        self.action[:self._curr_size] = data['action'][:self._curr_size]
        self.reward[:self._curr_size] = data['reward'][:self._curr_size]
        self.terminal[:self._curr_size] = data['terminal'][:self._curr_size]
        self.next_obs[:self._curr_size] = data['next_obs'][:self._curr_size]
        logger.info("[load rpm]memory loade from {}".format(pathname))



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from gym import spaces
from multiagent.multi_discrete import MultiDiscrete
from multiagent.environment import MultiAgentEnv
import multiagent.scenarios as scenarios


class MAenv(MultiAgentEnv):
    """ multiagent environment warppers for maddpg
    """

    def __init__(self, scenario_name):
        # load scenario from script
        scenario = scenarios.load(scenario_name + ".py").Scenario()
        # create world
        world = scenario.make_world()
        # initial multiagent environment
        super().__init__(world, scenario.reset_world, scenario.reward,
                         scenario.observation)
        self.obs_shape_n = [
            self.get_shape(self.observation_space[i]) for i in range(self.n)
        ]
        self.act_shape_n = [
            self.get_shape(self.action_space[i]) for i in range(self.n)
        ]

    def get_shape(self, input_space):
        """
        Args:
            input_space: environment space

        Returns:
            space shape
        """
        if (isinstance(input_space, spaces.Box)):
            if (len(input_space.shape) == 1):
                return input_space.shape[0]
            else:
                return input_space.shape
        elif (isinstance(input_space, spaces.Discrete)):
            return input_space.n
        elif (isinstance(input_space, MultiDiscrete)):
            return sum(input_space.high - input_space.low + 1)
        else:
            print('[Error] shape is {}, not Box or Discrete or MultiDiscrete'.
                  format(input_space.shape))
            raise NotImplementedError



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import parl
import torch
import math
from torch.distributions import Normal
import torch.nn.functional as F
from copy import deepcopy

device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

__all__ = ['OAC']


class OAC(parl.Algorithm):
    def __init__(self,
                 model,
                 gamma=None,
                 tau=None,
                 alpha=None,
                 beta=None,
                 delta=None,
                 actor_lr=None,
                 critic_lr=None):
        """ OAC algorithm
                    Args:
                        model(parl.Model): forward network of actor and critic.
                        gamma(float): discounted factor for reward computation
                        tau (float): decay coefficient when updating the weights of self.target_model with self.model
                        alpha (float): Temperature parameter determines the relative importance of the entropy against the reward
                        beta (float): determines the relative importance of sigma_Q
                        delta (float): determines the relative changes of exploration`s mean
                        actor_lr (float): learning rate of the actor model
                        critic_lr (float): learning rate of the critic model
        """
        assert isinstance(gamma, float)
        assert isinstance(tau, float)
        assert isinstance(alpha, float)
        assert isinstance(beta, float)
        assert isinstance(delta, float)
        assert isinstance(actor_lr, float)
        assert isinstance(critic_lr, float)

        self.gamma = gamma
        self.tau = tau
        self.alpha = alpha
        self.beta = beta
        self.delta = delta

        self.actor_lr = actor_lr
        self.critic_lr = critic_lr

        self.model = model.to(device)
        self.target_model = deepcopy(self.model)
        self.actor_optimizer = torch.optim.Adam(
            self.model.get_actor_params(), lr=actor_lr)
        self.critic_optimizer = torch.optim.Adam(
            self.model.get_critic_params(), lr=critic_lr)

    def predict(self, obs):
        act_mean, _ = self.model.policy(obs)
        action = torch.tanh(act_mean)
        return action

    def sample(self, obs):
        act_mean, act_log_std = self.model.policy(obs)
        normal = Normal(act_mean, act_log_std.exp())
        # for reparameterization trick  (mean + std*N(0,1))
        x_t = normal.rsample()
        action = torch.tanh(x_t)

        log_prob = normal.log_prob(x_t)
        log_prob -= torch.log((1 - action.pow(2)) + 1e-6)
        log_prob = log_prob.sum(1, keepdims=True)
        return action, log_prob

    def get_optimistic_exploration_action(self, obs):
        act_mean, act_log_std = self.model.policy(obs)
        act_std = torch.exp(act_log_std)
        normal = Normal(act_mean, act_std)
        pre_tanh_mu_T = normal.rsample()

        pre_tanh_mu_T.requires_grad_()
        tanh_mu_T = torch.tanh(pre_tanh_mu_T)

        # Get the upper bound of the Q estimate
        Q1, Q2 = self.model.critic_model(obs, tanh_mu_T)
        mu_Q = (Q1 + Q2) / 2.0
        sigma_Q = torch.abs(Q1 - Q2) / 2.0

        Q_UB = mu_Q + self.beta * sigma_Q

        # Obtain the gradient of Q_UB wrt to a with a evaluated at mu_t
        grad = torch.autograd.grad(Q_UB, pre_tanh_mu_T)
        grad = grad[0]
        assert grad is not None
        assert pre_tanh_mu_T.shape == grad.shape

        # Obtain Sigma_T (the covariance of the normal distribution)
        Sigma_T = torch.pow(act_std, 2)

        # The dividor is (g^T Sigma g) ** 0.5
        # Sigma is diagonal, so this works out to be
        # ( sum_{i=1}^k (g^(i))^2 (sigma^(i))^2 ) ** 0.5
        denom = torch.sqrt(torch.sum(torch.mul(torch.pow(grad, 2),
                                               Sigma_T))) + 10e-6

        # Obtain the change in mu
        mu_C = math.sqrt(2.0 * self.delta) * torch.mul(Sigma_T, grad) / denom
        assert mu_C.shape == pre_tanh_mu_T.shape

        mu_E = pre_tanh_mu_T + mu_C
        # Construct the tanh normal distribution and sample the exploratory action from it
        assert mu_E.shape == act_std.shape

        dist = Normal(mu_E, act_std)
        z = dist.sample().detach()
        action = torch.tanh(z)
        return action

    def learn(self, obs, action, reward, next_obs, terminal):
        critic_loss = self._critic_learn(obs, action, reward, next_obs,
                                         terminal)
        actor_loss = self._actor_learn(obs)

        self.sync_target()
        return critic_loss, actor_loss

    def _critic_learn(self, obs, action, reward, next_obs, terminal):
        with torch.no_grad():
            next_action, next_log_pro = self.sample(next_obs)
            q1_next, q2_next = self.target_model.value(next_obs, next_action)
            target_Q = torch.min(q1_next, q2_next) - self.alpha * next_log_pro
            target_Q = reward + self.gamma * (1. - terminal) * target_Q
        cur_q1, cur_q2 = self.model.critic_model(obs, action)

        critic_loss = F.mse_loss(cur_q1, target_Q) + F.mse_loss(
            cur_q2, target_Q)

        self.critic_optimizer.zero_grad()
        critic_loss.backward()
        self.critic_optimizer.step()
        return critic_loss

    def _actor_learn(self, obs):
        act, log_pi = self.sample(obs)
        q1_pi, q2_pi = self.model.critic_model(obs, act)
        min_q_pi = torch.min(q1_pi, q2_pi)
        actor_loss = ((self.alpha * log_pi) - min_q_pi).mean()

        self.actor_optimizer.zero_grad()
        actor_loss.backward()
        self.actor_optimizer.step()
        return actor_loss

    def sync_target(self, decay=None):
        if decay is None:
            decay = 1.0 - self.tau
        for param, target_param in zip(self.model.parameters(),
                                       self.target_model.parameters()):
            target_param.data.copy_((1 - decay) * param.data +
                                    decay * target_param.data)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings
warnings.simplefilter('default')

import paddle.fluid as fluid
from parl.core.fluid.algorithm import Algorithm
from parl.core.fluid import layers
from parl.algorithms.fluid.impala import vtrace
from parl.core.fluid.policy_distribution import CategoricalDistribution
from parl.core.fluid.plutils import inverse

__all__ = ['IMPALA']


class VTraceLoss(object):
    def __init__(self,
                 behaviour_actions_log_probs,
                 target_actions_log_probs,
                 policy_entropy,
                 dones,
                 discount,
                 rewards,
                 values,
                 bootstrap_value,
                 entropy_coeff=-0.01,
                 vf_loss_coeff=0.5,
                 clip_rho_threshold=1.0,
                 clip_pg_rho_threshold=1.0):
        """Policy gradient loss with vtrace importance weighting.

        VTraceLoss takes tensors of shape [T, B, ...], where `B` is the
        batch_size. The reason we need to know `B` is for V-trace to properly
        handle episode cut boundaries.

        Args:
            behaviour_actions_log_probs: A float32 tensor of shape [T, B].
            target_actions_log_probs: A float32 tensor of shape [T, B].
            policy_entropy: A float32 tensor of shape [T, B].
            dones: A float32 tensor of shape [T, B].
            discount: A float32 scalar.
            rewards: A float32 tensor of shape [T, B].
            values: A float32 tensor of shape [T, B].
            bootstrap_value: A float32 tensor of shape [B].
        """

        self.vtrace_returns = vtrace.from_importance_weights(
            behaviour_actions_log_probs=behaviour_actions_log_probs,
            target_actions_log_probs=target_actions_log_probs,
            discounts=inverse(dones) * discount,
            rewards=rewards,
            values=values,
            bootstrap_value=bootstrap_value,
            clip_rho_threshold=clip_rho_threshold,
            clip_pg_rho_threshold=clip_pg_rho_threshold)

        # The policy gradients loss
        self.pi_loss = -1.0 * layers.reduce_sum(
            target_actions_log_probs * self.vtrace_returns.pg_advantages)

        # The baseline loss
        delta = values - self.vtrace_returns.vs
        self.vf_loss = 0.5 * layers.reduce_sum(layers.square(delta))

        # The entropy loss (We want to maximize entropy, so entropy_ceoff < 0)
        self.entropy = layers.reduce_sum(policy_entropy)

        # The summed weighted loss
        self.total_loss = (self.pi_loss + self.vf_loss * vf_loss_coeff +
                           self.entropy * entropy_coeff)


class IMPALA(Algorithm):
    def __init__(self,
                 model,
                 sample_batch_steps=None,
                 gamma=None,
                 vf_loss_coeff=None,
                 clip_rho_threshold=None,
                 clip_pg_rho_threshold=None):
        r""" IMPALA algorithm
        
        Args:
            model (parl.Model): forward network of policy and value
            sample_batch_steps (int): steps of each environment sampling.
            gamma (float): discounted factor for reward computation.
            vf_loss_coeff (float): coefficient of the value function loss.
            clip_rho_threshold (float): clipping threshold for importance weights (rho).
            clip_pg_rho_threshold (float): clipping threshold on rho_s in \rho_s \delta log \pi(a|x) (r + \gamma v_{s+1} - V(x_s)).
        """
        assert isinstance(sample_batch_steps, int)
        assert isinstance(gamma, float)
        assert isinstance(vf_loss_coeff, float)
        assert isinstance(clip_rho_threshold, float)
        assert isinstance(clip_pg_rho_threshold, float)
        self.sample_batch_steps = sample_batch_steps
        self.gamma = gamma
        self.vf_loss_coeff = vf_loss_coeff
        self.clip_rho_threshold = clip_rho_threshold
        self.clip_pg_rho_threshold = clip_pg_rho_threshold

        self.model = model

    def learn(self, obs, actions, behaviour_logits, rewards, dones,
              learning_rate, entropy_coeff):
        """
        Args:
            obs: An float32 tensor of shape ([B] + observation_space).
                 E.g. [B, C, H, W] in atari.
            actions: An int64 tensor of shape [B].
            behaviour_logits: A float32 tensor of shape [B, NUM_ACTIONS].
            rewards: A float32 tensor of shape [B].
            dones: A float32 tensor of shape [B].
            learning_rate: float scalar of learning rate.
            entropy_coeff: float scalar of entropy coefficient.
        """

        values = self.model.value(obs)
        target_logits = self.model.policy(obs)

        target_policy_distribution = CategoricalDistribution(target_logits)
        behaviour_policy_distribution = CategoricalDistribution(
            behaviour_logits)

        policy_entropy = target_policy_distribution.entropy()
        target_actions_log_probs = target_policy_distribution.logp(actions)
        behaviour_actions_log_probs = behaviour_policy_distribution.logp(
            actions)

        # Calculating kl for debug
        kl = target_policy_distribution.kl(behaviour_policy_distribution)
        kl = layers.reduce_mean(kl)
        """
        Split the tensor into batches at known episode cut boundaries. 
        [B * T] -> [T, B]
        """
        T = self.sample_batch_steps

        def split_batches(tensor):
            B = tensor.shape[0] // T
            splited_tensor = layers.reshape(tensor,
                                            [B, T] + list(tensor.shape[1:]))
            # transpose B and T
            return layers.transpose(
                splited_tensor, [1, 0] + list(range(2, 1 + len(tensor.shape))))

        behaviour_actions_log_probs = split_batches(
            behaviour_actions_log_probs)
        target_actions_log_probs = split_batches(target_actions_log_probs)
        policy_entropy = split_batches(policy_entropy)
        dones = split_batches(dones)
        rewards = split_batches(rewards)
        values = split_batches(values)

        # [T, B] -> [T - 1, B] for V-trace calc.
        behaviour_actions_log_probs = layers.slice(
            behaviour_actions_log_probs, axes=[0], starts=[0], ends=[-1])
        target_actions_log_probs = layers.slice(
            target_actions_log_probs, axes=[0], starts=[0], ends=[-1])
        policy_entropy = layers.slice(
            policy_entropy, axes=[0], starts=[0], ends=[-1])
        dones = layers.slice(dones, axes=[0], starts=[0], ends=[-1])
        rewards = layers.slice(rewards, axes=[0], starts=[0], ends=[-1])
        bootstrap_value = layers.slice(
            values, axes=[0], starts=[T - 1], ends=[T])
        values = layers.slice(values, axes=[0], starts=[0], ends=[-1])

        bootstrap_value = layers.squeeze(bootstrap_value, axes=[0])

        vtrace_loss = VTraceLoss(
            behaviour_actions_log_probs=behaviour_actions_log_probs,
            target_actions_log_probs=target_actions_log_probs,
            policy_entropy=policy_entropy,
            dones=dones,
            discount=self.gamma,
            rewards=rewards,
            values=values,
            bootstrap_value=bootstrap_value,
            entropy_coeff=entropy_coeff,
            vf_loss_coeff=self.vf_loss_coeff,
            clip_rho_threshold=self.clip_rho_threshold,
            clip_pg_rho_threshold=self.clip_pg_rho_threshold)

        fluid.clip.set_gradient_clip(
            clip=fluid.clip.GradientClipByGlobalNorm(clip_norm=40.0))

        optimizer = fluid.optimizer.AdamOptimizer(learning_rate)
        optimizer.minimize(vtrace_loss.total_loss)
        return vtrace_loss, kl

    def sample(self, obs):
        """
        Args:
            obs: An float32 tensor of shape ([B] + observation_space).
                 E.g. [B, C, H, W] in atari.
        """
        logits = self.model.policy(obs)
        policy_dist = CategoricalDistribution(logits)
        sample_actions = policy_dist.sample()
        return sample_actions, logits

    def predict(self, obs):
        """
        Args:
            obs: An float32 tensor of shape ([B] + observation_space).
                 E.g. [B, C, H, W] in atari.
        """
        logits = self.model.policy(obs)
        probs = layers.softmax(logits)

        predict_actions = layers.argmax(probs, 1)

        return predict_actions



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

__all__ = ['WindowStat']

import numpy as np


class WindowStat(object):
    """ Tool to maintain statistical data in a window.
    """

    def __init__(self, window_size):
        self.items = [None] * window_size
        self.idx = 0
        self.count = 0

    def add(self, obj):
        self.items[self.idx] = obj
        self.idx += 1
        self.count += 1
        self.idx %= len(self.items)

    @property
    def mean(self):
        if self.count > 0:
            return np.mean(self.items[:self.count])
        else:
            return None

    @property
    def min(self):
        if self.count > 0:
            return np.min(self.items[:self.count])
        else:
            return None

    @property
    def max(self):
        if self.count > 0:
            return np.max(self.items[:self.count])
        else:
            return None



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import platform
import random
import socket
import subprocess
from parl.utils import logger, _HAS_FLUID, _HAS_PADDLE, _IS_WINDOWS

__all__ = [
    'get_gpu_count', 'get_ip_address', 'is_gpu_available', 'get_free_tcp_port',
    'is_port_available', 'get_port_from_range', 'is_xpu_available'
]


def get_ip_address():
    """
    get the IP address of the host.
    """

    # Windows
    if _IS_WINDOWS:
        local_ip = socket.gethostbyname(socket.gethostname())
    else:
        # Linux and MacOS
        local_ip = None
        try:
            # First way, tested in Ubuntu and MacOS
            s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            s.connect(("8.8.8.8", 80))
            local_ip = s.getsockname()[0]
            s.close()
        except:
            # Second way, tested in CentOS
            try:
                local_ip = socket.gethostbyname(socket.gethostname())
            except:
                pass

    if local_ip == None or local_ip == '127.0.0.1' or local_ip == '127.0.1.1':
        logger.warning(
            'get_ip_address failed, please set ip address manually.')
        return None

    return local_ip


def get_gpu_count():
    """get avaliable gpu count

    Returns:
        gpu_count: int
    """

    gpu_count = 0

    env_cuda_devices = os.environ.get('CUDA_VISIBLE_DEVICES', None)
    if env_cuda_devices is not None:
        assert isinstance(env_cuda_devices, str)
        try:
            if not env_cuda_devices:
                return 0
            gpu_count = len(
                [x for x in env_cuda_devices.split(',') if int(x) >= 0])
            logger.info(
                'CUDA_VISIBLE_DEVICES found gpu count: {}'.format(gpu_count))
        except:
            logger.info(
                'Cannot find available GPU devices, using CPU or other devices now.'
            )
            gpu_count = 0
    else:
        try:
            gpu_count = str(subprocess.check_output(["nvidia-smi",
                                                     "-L"])).count('UUID')
            logger.info('nvidia-smi -L found gpu count: {}'.format(gpu_count))
        except:
            logger.info(
                'Cannot find available GPU devices, using CPU or other devices now. (Please check whether you can execute `nvidia-smi` command.)'
            )
            gpu_count = 0
    return gpu_count


def is_gpu_available():
    """ check whether parl can access a GPU

    Returns:
      True if a gpu device can be found.
    """
    ret = get_gpu_count() > 0
    if _HAS_PADDLE:
        import paddle
        if ret is True and not paddle.is_compiled_with_cuda():
            logger.warning("Found non-empty CUDA_VISIBLE_DEVICES. \
                But PARL found that Paddle was not complied with CUDA, which may cause issues. \
                Thus PARL will not use GPU.")
            return False
    if _HAS_FLUID:
        from paddle import fluid
        if ret is True and not fluid.is_compiled_with_cuda():
            logger.warning("Found non-empty CUDA_VISIBLE_DEVICES. \
                But PARL found that Paddle was not complied with CUDA, which may cause issues. \
                Thus PARL will not use GPU.")
            return False
    return ret


def is_xpu_available():
    """ check whether parl can access a XPU

    Returns:
      True if paddle was complied with XPU.
    """
    xpu_count = int(os.getenv("FLAGS_selected_xpus", "-1"))
    if xpu_count < 0:
        return False

    if _HAS_FLUID:
        from paddle import fluid
        if not fluid.is_compiled_with_xpu():
            logger.warning("Found non-empty XPU_VISIBLE_DEVICES. \
                But PARL found that Paddle was not complied with XPU, which may cause issues. \
                Thus PARL will not use XPU.")
            return False
    if _HAS_PADDLE:
        import paddle
        if not paddle.is_compiled_with_xpu():
            logger.warning("Found non-empty XPU_VISIBLE_DEVICES. \
                But PARL found that Paddle was not complied with XPU, which may cause issues. \
                Thus PARL will not use XPU.")
            return False
    return True


def get_free_tcp_port():
    tcp = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    tcp.bind(('', 0))
    addr, port = tcp.getsockname()
    tcp.close()
    return str(port)


def is_port_available(port):
    """ Check if a port is used.

    True if the port is available for connection.
    """
    port = int(port)
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    available = sock.connect_ex(('localhost', port))
    sock.close()
    return available


def get_port_from_range(start, end):
    while True:
        port = random.randint(start, end)
        if is_port_available(port):
            break

    return port



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import parl
from parl.core.algorithm_base import AlgorithmBase

__all__ = ['Algorithm']


class Algorithm(AlgorithmBase):
    """
    | `alias`: ``parl.Algorithm``
    | `alias`: ``parl.core.fluid.algorithm.Algorithm``

    | ``Algorithm`` defines the way how to update the parameters of the ``Model``. This is where we define loss functions and the optimizer of the neural network. An ``Algorithm`` has at least a model.

    | PARL has implemented various algorithms(DQN/DDPG/PPO/A3C/IMPALA) that can be reused quickly, which can be accessed with ``parl.algorithms``.

    Example:

    .. code-block:: python

        import parl

        model = Model()
        dqn = parl.algorithms.DQN(model, lr=1e-3)

    Attributes:
        model(``parl.Model``): a neural network that represents a policy or a Q-value function.

    Pulic Functions:
        - ``get_weights``: return a Python dictionary containing parameters of the current model.
        - ``set_weights``: copy parameters from ``get_weights()`` to the model.
        - ``sample``: return a noisy action to perform exploration according to the policy.
        - ``predict``: return an action given current observation.
        - ``learn``: define the loss function and create an optimizer to minized the loss.

    Note:

        ``Algorithm`` defines all its computation inside a ``fluid.Program``, such that the returns of functions(`sample`, `predict`, `learn`) are tensors.
        ``Agent`` also has functions like `sample`, `predict`, and `learn`, but they return numpy array for the agent.
        
    """

    def __init__(self, model=None):
        """
        Args:
            model(``parl.Model``): a neural network that represents a policy or a Q-value function.
        """
        assert isinstance(model, parl.Model)
        self.model = model

    def get_weights(self, model_ids=None):
        """Get weights of all `parl.Model` in self.__dict__.
        
        If `model_ids` is not None, will only return weights of
        models whose model_id are in `model_ids`.
        
        Note:
            `parl.Model` in list, tuple and dict will be included. But `parl.Model` in
            nested list, tuple and dict won't be included.

        Args:
            model_ids (List/Set): list/set of model_id, will only return weights of models
                              whose model_id in the `model_ids`.

        Returns:
            Dict of weights ({attribute name: numpy array/List/Dict})
        """
        if model_ids is not None:
            assert isinstance(model_ids, (list, set))
            model_ids = set(model_ids)

        model_weights = {}
        for key in self.__dict__.keys():
            value = getattr(self, key)
            if isinstance(value, parl.Model):
                if model_ids is None or value.model_id in model_ids:
                    model_weights[key] = value.get_weights()
            elif isinstance(value, list) or isinstance(value, tuple):
                weights_list = []
                for x in value:
                    if isinstance(x, parl.Model):
                        if model_ids is None or x.model_id in model_ids:
                            weights_list.append(x.get_weights())
                if weights_list:
                    model_weights[key] = weights_list
            elif isinstance(value, dict):
                weights_dict = {}
                for sub_k, sub_v in value.items():
                    if isinstance(sub_v, parl.Model):
                        if model_ids is None or sub_v.model_id in model_ids:
                            weights_dict[sub_k] = sub_v.get_weights()
                if weights_dict:
                    model_weights[key] = weights_dict
        return model_weights

    def set_weights(self, weights, model_ids=None):
        """Set weights of all `parl.Model` in self.__dict__.

        If `model_ids` is not None, will only set weights of
        models whose model_id are in `model_ids`.

        Note:
            `parl.Model` in list, tuple and dict will be included. But `parl.Model` in
            nested list, tuple and dict won't be included.

        Args:
            weights (Dict): Dict of weights ({attribute name: numpy array/List/Dict})
            model_ids (List/Set): list/set of model_id, will only set weights of models
                              whiose model_id in the `model_ids`.
        
        """
        assert isinstance(weights, dict)
        if model_ids is not None:
            assert isinstance(model_ids, (list, set))
            model_ids = set(model_ids)

        for key in self.__dict__.keys():
            value = getattr(self, key)
            if isinstance(value, parl.Model):
                if model_ids is None or value.model_id in model_ids:
                    assert key in weights, "weights is inconsistent with current algorithm and given model_ids."
                    value.set_weights(weights[key])
            elif isinstance(value, list) or isinstance(value, tuple):
                model_list = []
                for x in value:
                    if isinstance(x, parl.Model):
                        if model_ids is None or x.model_id in model_ids:
                            model_list.append(x)
                if model_list:
                    assert key in weights and len(model_list) == len(weights[key]), \
                            "weights is inconsistent with current algorithm and given model_ids."
                    for i, model in enumerate(model_list):
                        model.set_weights(weights[key][i])
            elif isinstance(value, dict):
                model_dict = {}
                for sub_k, sub_v in value.items():
                    if isinstance(sub_v, parl.Model):
                        if model_ids is None or sub_v.model_id in model_ids:
                            model_dict[sub_k] = sub_v
                if model_dict:
                    assert key in weights and set(model_dict.keys()) == set(weights[key].keys()), \
                            "weights is inconsistent with current algorithm and given model_ids."
                    for sub_k, model in model_dict.items():
                        model.set_weights(weights[key][sub_k])

    def get_model_ids(self):
        """Get model_id of all `parl.Model` in self.__dict__.

        Note:
            `parl.Model` in list, tuple and dict will be included. But `parl.Model` in
            nested list, tuple and dict won't be included.
        
        Returns:
            Set of model_id 
        """
        model_ids = set([])
        for key in self.__dict__.keys():
            value = getattr(self, key)
            if isinstance(value, parl.Model):
                model_ids.add(value.model_id)
            elif isinstance(value, list) or isinstance(value, tuple):
                for x in value:
                    if isinstance(x, parl.Model):
                        model_ids.add(x.model_id)
            elif isinstance(value, dict):
                for sub_k, sub_v in value.items():
                    if isinstance(sub_v, parl.Model):
                        model_ids.add(sub_v.model_id)
        return model_ids

    @property
    def model_ids(self):
        return self.get_model_ids()

    def learn(self, *args, **kwargs):
        """ Define the loss function and create an optimizer to minize the loss.
        """
        raise NotImplementedError

    def predict(self, *args, **kwargs):
        """ Refine the predicting process, e.g,. use the policy model to predict actions.
        """
        raise NotImplementedError

    def sample(self, *args, **kwargs):
        """ Define the sampling process. This function returns an action with noise to perform exploration.
        """
        raise NotImplementedError



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class UtilsError(Exception):
    """
    Super class of exceptions in utils module.
    """

    def __init__(self, error_info):
        self.error_info = '[PARL Utils Error]: {}'.format(error_info)


class SerializeError(UtilsError):
    """
    Serialize error raised by pyarrow.
    """

    def __init__(self, error_info):
        error_info = (
            'Serialize error, you may have provided an object that cannot be '
            + 'serialized by pyarrow. Detailed error:\n{}'.format(error_info))
        super(SerializeError, self).__init__(error_info)

    def __str__(self):
        return self.error_info


class DeserializeError(UtilsError):
    """
    Deserialize error raised by pyarrow.
    """

    def __init__(self, error_info):
        error_info = (
            'Deserialize error, you may have provided an object that cannot be '
            +
            'deserialized by pyarrow. Detailed error:\n{}'.format(error_info))
        super(DeserializeError, self).__init__(error_info)

    def __str__(self):
        return self.error_info



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np

__all__ = ['np_softmax', 'np_cross_entropy']


def np_softmax(logits):
    return np.exp(logits) / np.sum(np.exp(logits), axis=-1, keepdims=True)


def np_cross_entropy(probs, labels):
    if labels.shape[-1] == 1:
        # sparse label
        n_classes = probs.shape[-1]
        result_shape = list(labels.shape[:-1]) + [n_classes]
        labels = np.eye(n_classes)[labels.reshape(-1)]
        labels = labels.reshape(result_shape)

    return -np.sum(labels * np.log(probs), axis=-1, keepdims=True)



#   Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import torch
import os
from copy import deepcopy
import parl
import numpy as np

device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

__all__ = ['COMA']


class COMA(parl.Algorithm):
    def __init__(self,
                 model,
                 n_actions,
                 n_agents,
                 grad_norm_clip=None,
                 actor_lr=None,
                 critic_lr=None,
                 gamma=None,
                 td_lambda=None):
        """  COMA algorithm
        
        Args:
            model (parl.Model): forward network of actor and critic.
            n_actions (int): action dim for each agent
            n_agents (int): agents number
            grad_norm_clip (int or float): gradient clip, prevent gradient explosion
            actor_lr (float): actor network learning rate
            critic_lr (float): critic network learning rate
            gamma (float):  discounted factor for reward computation
            td_lambda (float): lambda of td-lambda return
        """
        assert isinstance(n_actions, int)
        assert isinstance(n_agents, int)
        assert isinstance(grad_norm_clip, int) or isinstance(
            grad_norm_clip, float)
        assert isinstance(actor_lr, float)
        assert isinstance(critic_lr, float)
        assert isinstance(gamma, float)
        assert isinstance(td_lambda, float)

        self.n_actions = n_actions
        self.n_agents = n_agents
        self.grad_norm_clip = grad_norm_clip
        self.actor_lr = actor_lr
        self.critic_lr = critic_lr
        self.gamma = gamma
        self.td_lambda = td_lambda

        self.model = model.to(device)
        self.target_model = deepcopy(model).to(device)

        self.sync_target()

        self.actor_parameters = list(self.model.get_actor_params())
        self.critic_parameters = list(self.model.get_critic_params())

        self.critic_optimizer = torch.optim.RMSprop(
            self.critic_parameters, lr=self.critic_lr)
        self.actor_optimizer = torch.optim.RMSprop(
            self.actor_parameters, lr=self.actor_lr)

        self.train_rnn_h = None

    def init_hidden(self, ep_num):
        """ function: init a hidden tensor for every agent
            input: 
                ep_num: How many episodes are included in a batch of data
            output:
                rnn_h: rnn hidden state, shape (ep_num, n_agents, hidden_size)
        """
        assert hasattr(self.model.actor_model, 'init_hidden'), \
            "actor must have rnn structure and has method 'init_hidden' to make hidden states"
        rnn_h = self.model.actor_model.init_hidden().unsqueeze(0).expand(
            ep_num, self.n_agents, -1)
        return rnn_h

    def predict(self, obs, rnn_h_in):
        """input:
                obs: obs + last_action + agent_id, shape: (1, obs_shape + n_actions + n_agents)
                rnn_h_in: rnn's hidden input
            output:
                prob: output of actor, shape: (1, n_actions)
                rnn_h_out: rnn's hidden output
        """
        with torch.no_grad():
            policy_logits, rnn_h_out = self.model.policy(
                obs, rnn_h_in)  # input obs shape [1, 42]
            prob = torch.nn.functional.softmax(
                policy_logits, dim=-1)  # shape [1, 9]
        return prob, rnn_h_out

    def _get_critic_output(self, batch):
        """ input:
                batch: dict(o, s, u, r, u_onehot, avail_u, padded, isover, actor_inputs, critic_inputs)
            output:
                q_evals and q_targets: shape (ep_num, tr_num, n_agents, n_actions)
        """
        ep_num = batch['r'].shape[0]
        tr_num = batch['r'].shape[1]
        critic_inputs = batch['critic_inputs']
        critic_inputs_next = batch['critic_inputs_next']

        critic_inputs = critic_inputs.reshape((ep_num * tr_num * self.n_agents,
                                               -1))
        critic_inputs_next = critic_inputs.reshape(
            (ep_num * tr_num * self.n_agents, -1))

        q_evals = self.model.value(critic_inputs)
        q_targets = self.model.value(critic_inputs_next)

        q_evals = q_evals.reshape((ep_num, tr_num, self.n_agents, -1))
        q_targets = q_targets.reshape((ep_num, tr_num, self.n_agents, -1))
        return q_evals, q_targets

    def _get_actor_output(self, batch, epsilon):
        """ input:
                batch: dict(o, s, u, r, u_onehot, avail_u, padded, isover, actor_inputs, critic_inputs)
                epsilon: noise discount factor
            output:
                action_prob: probability of actions, shape (ep_num, tr_num, n_agents, n_actions)
        """
        ep_num = batch['r'].shape[0]
        tr_num = batch['r'].shape[1]
        avail_actions = batch['avail_u']
        actor_inputs = batch['actor_inputs']
        action_prob = []
        for tr_id in range(tr_num):
            inputs = actor_inputs[:,
                                  tr_id]  # shape (ep_num, n_agents, actor_input_dim)
            inputs = inputs.reshape(
                (-1, inputs.shape[-1]))  # shape (-1, actor_input_dim)
            policy_logits, self.train_rnn_h = self.model.policy(
                inputs, self.train_rnn_h)
            # policy_logits shape from (-1, n_actions) to (ep_num, n_agents, n_actions)
            policy_logits = policy_logits.view(ep_num, self.n_agents, -1)
            prob = torch.nn.functional.softmax(policy_logits, dim=-1)
            action_prob.append(prob)

        action_prob = torch.stack(
            action_prob,
            dim=1).to(device)  # shape: (ep_num, tr_num, n_agents, n_actions)
        action_num = avail_actions.sum()  # how many actions are available
        action_prob = ((1 - epsilon) * action_prob +
                       torch.ones_like(action_prob) * epsilon / action_num)
        action_prob[avail_actions == 0] = 0.0  # set avail action

        action_prob = action_prob / action_prob.sum(
            dim=-1, keepdim=True)  # in case action_prob.sum != 1
        action_prob[avail_actions == 0] = 0.0
        action_prob = action_prob.to(device)
        return action_prob

    def _cal_td_target(self, batch, q_targets):  # compute TD(lambda)
        """ input:
                batch: dict(o, s, u, r, u_onehot, avail_u, padded, isover, actor_inputs, critic_inputs)
                q_targets: Q value of target critic network, shape (ep_num, tr_num, n_agents)
            output:
                lambda_return: TD lambda return, shape (ep_num, tr_num, n_agents)
        """
        ep_num = batch['r'].shape[0]
        tr_num = batch['r'].shape[1]
        mask = (1 - batch['padded'].float()).repeat(1, 1,
                                                    self.n_agents).to(device)
        isover = (1 - batch['isover'].float()).repeat(1, 1, self.n_agents).to(
            device)  # used for setting last transition's q_target to 0
        # reshape reward: from (ep_num, tr_num, 1) to (ep_num, tr_num, n_agents)
        r = batch['r'].repeat((1, 1, self.n_agents)).to(device)
        # compute n_step_return
        n_step_return = torch.zeros((ep_num, tr_num, self.n_agents,
                                     tr_num)).to(device)
        for tr_id in range(tr_num - 1, -1, -1):
            n_step_return[:, tr_id, :, 0] = (
                r[:, tr_id] + self.gamma * q_targets[:, tr_id] *
                isover[:, tr_id]) * mask[:, tr_id]
            for n in range(1, tr_num - tr_id):
                n_step_return[:, tr_id, :, n] = (
                    r[:, tr_id] + self.gamma *
                    n_step_return[:, tr_id + 1, :, n - 1]) * mask[:, tr_id]

        lambda_return = torch.zeros((ep_num, tr_num, self.n_agents)).to(device)
        for tr_id in range(tr_num):
            returns = torch.zeros((ep_num, self.n_agents)).to(device)
            for n in range(1, tr_num - tr_id):
                returns += pow(self.td_lambda,
                               n - 1) * n_step_return[:, tr_id, :, n - 1]
            lambda_return[:, tr_id] = (1 - self.td_lambda) * returns + \
                                            pow(self.td_lambda, tr_num - tr_id - 1) * \
                                            n_step_return[:, tr_id, :, tr_num - tr_id - 1]
        return lambda_return

    def _critic_learn(self, batch):
        """ input:
                batch: dict(o, s, u, r, u_onehot, avail_u, padded, isover, actor_inputs, critic_inputs)
            output:
                q_values: Q value of eval critic network, shape (ep_num, tr_num, n_agents, n_actions)
        """
        u = batch['u']  # shape (ep_num, tr_num, agent, n_actions)
        u_next = torch.zeros_like(u, dtype=torch.long)
        u_next[:, :-1] = u[:, 1:]
        mask = (1 - batch['padded'].float()).repeat(1, 1,
                                                    self.n_agents).to(device)

        # get q value for every agent and every action, shape (ep_num, tr_num, n_agents, n_actions)
        q_evals, q_next_target = self._get_critic_output(batch)
        q_values = q_evals.clone()  # used for function return

        # get q valur for every agent
        q_evals = torch.gather(q_evals, dim=3, index=u).squeeze(3)
        q_next_target = torch.gather(
            q_next_target, dim=3, index=u_next).squeeze(3)

        targets = self._cal_td_target(batch, q_next_target)

        td_error = targets.detach() - q_evals
        masked_td_error = mask * td_error  # mask padded data

        loss = (masked_td_error**
                2).sum() / mask.sum()  # mask.sum: avail transition num

        self.critic_optimizer.zero_grad()
        loss.backward()
        torch.nn.utils.clip_grad_norm_(self.critic_parameters,
                                       self.grad_norm_clip)
        self.critic_optimizer.step()
        return q_values

    def _actor_learn(self, batch, epsilon, q_values):
        """ input:
                batch: dict(o, s, u, r, u_onehot, avail_u, padded, isover, actor_inputs, critic_inputs)
                epsilon (float): e-greedy discount
                q_values: Q value of eval critic network, shape (ep_num, tr_num, n_agents, n_actions)
        """
        action_prob = self._get_actor_output(batch, epsilon)  # prob of u

        # mask: used to compute TD-error, filling data should not affect learning
        u = batch['u']
        mask = (1 - batch['padded'].float()).repeat(1, 1, self.n_agents).to(
            device)  # shape (ep_num, tr_num, 3)

        q_taken = torch.gather(q_values, dim=3, index=u).squeeze(3)  # Q(u_a)
        pi_taken = torch.gather(
            action_prob, dim=3,
            index=u).squeeze(3)  # prob of act that agent a choosen
        pi_taken[mask == 0] = 1.0  # prevent log overflow
        log_pi_taken = torch.log(pi_taken)

        # advantage
        baseline = (q_values * action_prob).sum(
            dim=3, keepdim=True).squeeze(3).detach()
        advantage = (q_taken - baseline).detach()
        loss = -((advantage * log_pi_taken) * mask).sum() / mask.sum()
        self.actor_optimizer.zero_grad()
        loss.backward()
        torch.nn.utils.clip_grad_norm_(self.actor_parameters,
                                       self.grad_norm_clip)
        self.actor_optimizer.step()

    def learn(self, batch, epsilon):
        """ input:
                batch: dict(o, s, u, r, u_onehot, avail_u, padded, isover, actor_inputs, critic_inputs)
                epsilon (float): e-greedy discount
        """
        ep_num = batch['r'].shape[0]
        self.train_rnn_h = self.init_hidden(ep_num)
        self.train_rnn_h = self.train_rnn_h.to(device)

        q_values = self._critic_learn(batch)
        self._actor_learn(batch, epsilon, q_values)

    def sync_target(self, decay=0):
        for param, target_param in zip(self.model.parameters(),
                                       self.target_model.parameters()):
            target_param.data.copy_((1 - decay) * param.data +
                                    decay * target_param.data)



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Reference:
    https://github.com/briancurtin/deprecation
"""

__all__ = ['deprecated']

import functools
import textwrap
import warnings

warnings.simplefilter('default')


class CustomDeprecationWarning(DeprecationWarning):
    def __init__(self,
                 function,
                 deprecated_in,
                 removed_in,
                 replace_function=None):
        """ 
        Args:
            function (String): The function being deprecated.
            deprecated_in (String): The version that ``function`` is deprecated in
            removed_in (String): The version that ``function`` gets removed in
            replace_function (String): The replacement function of deprecated function.
        """

        self.function = function
        self.deprecated_in = deprecated_in
        self.removed_in = removed_in
        self.replace_function = replace_function
        super(CustomDeprecationWarning, self).__init__(
            function, deprecated_in, removed_in, replace_function)

    def __str__(self):
        warn_string = '[PARL] API `{}` is deprecated since version {} and will be removed in version {}'.format(
            self.function, self.deprecated_in, self.removed_in)
        if self.replace_function is not None:
            warn_string += ", please use `{}` instead.".format(
                self.replace_function)
        else:
            warn_string += "."
        return warn_string


def deprecated(deprecated_in, removed_in, replace_function=None):
    """Decorator of decarated function.

    Args:
        function (String): The function being deprecated.
        deprecated_in (String): The version that ``function`` is deprecated in
        removed_in (String): The version that ``function`` gets removed in
        replace_function (String): The replacement function of deprecated function.
    """

    def _function_wrapper(function):
        existing_docstring = function.__doc__ or ""

        deprecated_doc = '.. deprecated:: {}\n    This will be removed in {}'.format(
            deprecated_in, removed_in)
        if replace_function is not None:
            deprecated_doc += ", please use `{}` instead.".format(
                replace_function)
        else:
            deprecated_doc += "."

        # split docstring at first occurrence of newline
        string_list = existing_docstring.split("\n", 1)

        if len(string_list) > 1:
            # in-place dedent docstring content
            string_list[1] = textwrap.dedent(string_list[1])

            # we need another newline
            string_list.insert(1, "\n")

        # insert deprecation note and dual newline
        string_list.insert(1, deprecated_doc)
        string_list.insert(1, "\n\n")

        function.__doc__ = "".join(string_list)

        @functools.wraps(function)
        def _inner(*args, **kwargs):
            the_warning = CustomDeprecationWarning(
                function.__name__, deprecated_in, removed_in, replace_function)
            warnings.warn(
                the_warning, category=CustomDeprecationWarning, stacklevel=2)

            return function(*args, **kwargs)

        return _inner

    return _function_wrapper



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from parl.core.fluid import layers

__all__ = [
    'PolicyDistribution', 'CategoricalDistribution',
    'SoftCategoricalDistribution', 'SoftMultiCategoricalDistribution'
]


class PolicyDistribution(object):
    def sample(self):
        """Sampling from the policy distribution."""
        raise NotImplementedError

    def entropy(self):
        """The entropy of the policy distribution."""
        raise NotImplementedError

    def kl(self, other):
        """The KL-divergence between self policy distributions and other."""
        raise NotImplementedError

    def logp(self, actions):
        """The log-probabilities of the actions in this policy distribution."""
        raise NotImplementedError


class CategoricalDistribution(PolicyDistribution):
    """Categorical distribution for discrete action spaces."""

    def __init__(self, logits):
        """
        Args:
            logits: A float32 tensor with shape [BATCH_SIZE, NUM_ACTIONS] of unnormalized policy logits
        """
        assert len(logits.shape) == 2
        self.logits = logits

    def sample(self):
        """
        Returns:
            sample_action: An int64 tensor with shape [BATCH_SIZE] of multinomial sampling ids.
                           Each value in sample_action is in [0, NUM_ACTIOINS - 1]
        """
        probs = layers.softmax(self.logits)
        sample_actions = layers.sampling_id(probs)
        return sample_actions

    def entropy(self):
        """
        Returns:
            entropy: A float32 tensor with shape [BATCH_SIZE] of entropy of self policy distribution.
        """
        logits = self.logits - layers.reduce_max(self.logits, dim=1)
        e_logits = layers.exp(logits)
        z = layers.reduce_sum(e_logits, dim=1)
        prob = e_logits / z
        entropy = -1.0 * layers.reduce_sum(
            prob * (logits - layers.log(z)), dim=1)

        return entropy

    def logp(self, actions, eps=1e-6):
        """
        Args:
            actions: An int64 tensor with shape [BATCH_SIZE]
            eps: A small float constant that avoids underflows when computing the log probability

        Returns:
            actions_log_prob: A float32 tensor with shape [BATCH_SIZE]
        """
        assert len(actions.shape) == 1

        logits = self.logits - layers.reduce_max(self.logits, dim=1)
        e_logits = layers.exp(logits)
        z = layers.reduce_sum(e_logits, dim=1)
        prob = e_logits / z

        actions = layers.unsqueeze(actions, axes=[1])
        actions_onehot = layers.one_hot(actions, prob.shape[1])
        actions_onehot = layers.cast(actions_onehot, dtype='float32')
        actions_prob = layers.reduce_sum(prob * actions_onehot, dim=1)

        actions_prob = actions_prob + eps
        actions_log_prob = layers.log(actions_prob)

        return actions_log_prob

    def kl(self, other):
        """
        Args:
            other: object of CategoricalDistribution

        Returns:
            kl: A float32 tensor with shape [BATCH_SIZE]
        """
        assert isinstance(other, CategoricalDistribution)

        logits = self.logits - layers.reduce_max(self.logits, dim=1)
        other_logits = other.logits - layers.reduce_max(other.logits, dim=1)

        e_logits = layers.exp(logits)
        other_e_logits = layers.exp(other_logits)

        z = layers.reduce_sum(e_logits, dim=1)
        other_z = layers.reduce_sum(other_e_logits, dim=1)

        prob = e_logits / z
        kl = layers.reduce_sum(
            prob *
            (logits - layers.log(z) - other_logits + layers.log(other_z)),
            dim=1)
        return kl


class SoftCategoricalDistribution(CategoricalDistribution):
    """Categorical distribution with noise for discrete action spaces"""

    def __init__(self, logits):
        """
        Args:
            logits: A float32 tensor with shape [BATCH_SIZE, NUM_ACTIONS] of unnormalized policy logits
        """
        self.logits = logits
        super(SoftCategoricalDistribution, self).__init__(logits)

    def sample(self):
        """
        Returns:
            sample_action: An int64 tensor with shape [BATCH_SIZE, NUM_ACTIOINS] of sample action,
                           with noise to keep the target close to the original action.
        """
        eps = 1e-4
        logits_shape = layers.cast(layers.shape(self.logits), dtype='int64')
        uniform = layers.uniform_random(logits_shape, min=eps, max=1.0 - eps)
        soft_uniform = layers.log(-1.0 * layers.log(uniform))
        return layers.softmax(self.logits - soft_uniform, axis=-1)


class SoftMultiCategoricalDistribution(PolicyDistribution):
    """Categorical distribution with noise for MultiDiscrete action spaces."""

    def __init__(self, logits, low, high):
        """
        Args:
            logits: A float32 tensor with shape [BATCH_SIZE, LEN_MultiDiscrete, NUM_ACTIONS] of unnormalized policy logits
            low: lower bounds of sample action
            high: Upper bounds of action
        """
        self.logits = logits
        self.low = low
        self.high = high
        self.categoricals = list(
            map(
                SoftCategoricalDistribution,
                layers.split(
                    input=logits,
                    num_or_sections=list(high - low + 1),
                    dim=len(logits.shape) - 1)))

    def sample(self):
        """
        Returns:
            sample_action: An int64 tensor with shape [BATCH_SIZE, NUM_ACTIOINS] of sample action,
                           with noise to keep the target close to the original action.
        """
        cate_list = []
        for i in range(len(self.categoricals)):
            cate_list.append(self.low[i] + self.categoricals[i].sample())
        return layers.concat(cate_list, axis=-1)

    def layers_add_n(self, input_list):
        """
        Adds all input tensors element-wise, can replace tf.add_n
        """
        assert len(input_list) >= 1
        res = input_list[0]
        for i in range(1, len(input_list)):
            res = layers.elementwise_add(res, input_list[i])
        return res

    def entropy(self):
        """
        Returns:
            entropy: A float32 tensor with shape [BATCH_SIZE] of entropy of self policy distribution.
        """
        return self.layers_add_n([p.entropy() for p in self.categoricals])

    def kl(self, other):
        """
        Args:
            other: object of SoftCategoricalDistribution

        Returns:
            kl: A float32 tensor with shape [BATCH_SIZE]
        """
        return self.layers_add_n(
            [p.kl(q) for p, q in zip(self.categoricals, other.categoricals)])



#   Copyright (c) 2019 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

CPU_TAG = b'[CPU]'
CONNECT_TAG = b'[CONNECT]'
HEARTBEAT_TAG = b'[HEARTBEAT]'
KILLJOB_TAG = b'[KILLJOB]'
MONITOR_TAG = b'[MONITOR]'
STATUS_TAG = b'[STATUS]'

WORKER_CONNECT_TAG = b'[WORKER_CONNECT]'
WORKER_INITIALIZED_TAG = b'[WORKER_INITIALIZED]'
WORKER_STATUS_UPDATE_TAG = b'[WORKER_STATUS_UPDATE_TAG]'
CLIENT_CONNECT_TAG = b'[CLIENT_CONNECT]'
CLIENT_SUBMIT_TAG = b'[CLIENT_SUBMIT]'
CLIENT_STATUS_UPDATE_TAG = b'[CLIENT_STATUS_UPDATE_TAG]'
SEND_FILE_TAG = b'[SEND_FILE]'
SUBMIT_JOB_TAG = b'[SUBMIT_JOB]'
NEW_JOB_TAG = b'[NEW_JOB]'

CHECK_VERSION_TAG = b'[CHECK_VERSION]'
INIT_OBJECT_TAG = b'[INIT_OBJECT]'
CALL_TAG = b'[CALL]'
GET_ATTRIBUTE_TAG = b'[GET_ATTRIBUTE]'
SET_ATTRIBUTE_TAG = b'[SET_ATTRIBUTE]'

EXCEPTION_TAG = b'[EXCEPTION]'
ATTRIBUTE_EXCEPTION_TAG = b'[ATTRIBUTE_EXCEPTION]'
SERIALIZE_EXCEPTION_TAG = b'[SERIALIZE_EXCEPTION]'
DESERIALIZE_EXCEPTION_TAG = b'[DESERIALIZE_EXCEPTION]'

NORMAL_TAG = b'[NORMAL]'

# interval of heartbeat mechanism in the unit of second
HEARTBEAT_INTERVAL_S = 10
HEARTBEAT_TIMEOUT_S = 10
HEARTBEAT_RCVTIMEO_S = HEARTBEAT_INTERVAL_S + HEARTBEAT_TIMEOUT_S * 2

HEARTBEAT_OUT_OF_MEMORY_TAG = b'[HEARTBEAT_OUT_OF_MEMORY]'



#   Copyright (c) 2018 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions to compute V-trace off-policy actor critic targets,
which used in IMAPLA algorithm.

The following code is mainly referenced and copied from:
https://github.com/deepmind/scalable_agent/blob/master/vtrace.py

For details and theory see:

"Espeholt L, Soyer H, Munos R, et al. Impala: Scalable distributed 
deep-rl with importance weighted actor-learner 
architectures[J]. arXiv preprint arXiv:1802.01561, 2018."

"""

import collections
import paddle.fluid as fluid
from parl.core.fluid import layers
from parl.utils import MAX_INT32

VTraceReturns = collections.namedtuple('VTraceReturns',
                                       ['vs', 'pg_advantages'])


def from_importance_weights(behaviour_actions_log_probs,
                            target_actions_log_probs,
                            discounts,
                            rewards,
                            values,
                            bootstrap_value,
                            clip_rho_threshold=1.0,
                            clip_pg_rho_threshold=1.0,
                            name='vtrace_from_logits'):
    r"""V-trace for softmax policies.

    Calculates V-trace actor critic targets for softmax polices as described in

    "IMPALA: Scalable Distributed Deep-RL with
    Importance Weighted Actor-Learner Architectures"
    by Espeholt, Soyer, Munos et al.

    Target policy refers to the policy we are interested in improving and
    behaviour policy refers to the policy that generated the given
    rewards and actions.

    In the notation used throughout documentation and comments, T refers to the
    time dimension ranging from 0 to T-1. B refers to the batch size and
    NUM_ACTIONS refers to the number of actions.

    Args:
      behaviour_actions_log_probs: A float32 tensor of shape [T, B] of
        log-probabilities of actions in behaviour policy.
      target_policy_logits: A float32 tensor of shape [T, B] of
        log-probabilities of actions in target policy.
      discounts: A float32 tensor of shape [T, B] with the discount encountered
        when following the behaviour policy.
      rewards: A float32 tensor of shape [T, B] with the rewards generated by
        following the behaviour policy.
      values: A float32 tensor of shape [T, B] with the value function estimates
        wrt. the target policy.
      bootstrap_value: A float32 of shape [B] with the value function estimate at
        time T.
      clip_rho_threshold: A scalar float32 tensor with the clipping threshold for
        importance weights (rho) when calculating the baseline targets (vs).
        rho^bar in the paper.
      clip_pg_rho_threshold: A scalar float32 tensor with the clipping threshold
        on rho_s in \rho_s \delta log \pi(a|x) (r + \gamma v_{s+1} - V(x_s)).
      name: The name scope that all V-trace operations will be created in.

    Returns:
      A VTraceReturns namedtuple (vs, pg_advantages) where:
        vs: A float32 tensor of shape [T, B]. Can be used as target to
          train a baseline (V(x_t) - vs_t)^2.
        pg_advantages: A float32 tensor of shape [T, B]. Can be used as the
          advantage in the calculation of policy gradients.
    """

    rank = len(behaviour_actions_log_probs.shape)  # Usually 2.
    assert len(target_actions_log_probs.shape) == rank
    assert len(values.shape) == rank
    assert len(bootstrap_value.shape) == (rank - 1)
    assert len(discounts.shape) == rank
    assert len(rewards.shape) == rank

    # log importance sampling weights.
    # V-trace performs operations on rhos in log-space for numerical stability.
    log_rhos = target_actions_log_probs - behaviour_actions_log_probs

    if clip_rho_threshold is not None:
        clip_rho_threshold = layers.fill_constant([1], 'float32',
                                                  clip_rho_threshold)
    if clip_pg_rho_threshold is not None:
        clip_pg_rho_threshold = layers.fill_constant([1], 'float32',
                                                     clip_pg_rho_threshold)

    rhos = layers.exp(log_rhos)
    if clip_rho_threshold is not None:
        clipped_rhos = layers.elementwise_min(rhos, clip_rho_threshold)
    else:
        clipped_rhos = rhos

    constant_one = layers.fill_constant([1], 'float32', 1.0)
    cs = layers.elementwise_min(rhos, constant_one)

    # Append bootstrapped value to get [v1, ..., v_t+1]
    values_1_t = layers.slice(values, axes=[0], starts=[1], ends=[MAX_INT32])
    values_t_plus_1 = layers.concat(
        [values_1_t, layers.unsqueeze(bootstrap_value, [0])], axis=0)

    # \delta_s * V
    deltas = clipped_rhos * (rewards + discounts * values_t_plus_1 - values)

    vs_minus_v_xs = recursively_scan(discounts, cs, deltas)

    # Add V(x_s) to get v_s.
    vs = layers.elementwise_add(vs_minus_v_xs, values)

    # Advantage for policy gradient.
    vs_1_t = layers.slice(vs, axes=[0], starts=[1], ends=[MAX_INT32])
    vs_t_plus_1 = layers.concat(
        [vs_1_t, layers.unsqueeze(bootstrap_value, [0])], axis=0)

    if clip_pg_rho_threshold is not None:
        clipped_pg_rhos = layers.elementwise_min(rhos, clip_pg_rho_threshold)
    else:
        clipped_pg_rhos = rhos
    pg_advantages = (
        clipped_pg_rhos * (rewards + discounts * vs_t_plus_1 - values))

    # Make sure no gradients backpropagated through the returned values.
    vs.stop_gradient = True
    pg_advantages.stop_gradient = True
    return VTraceReturns(vs=vs, pg_advantages=pg_advantages)


def recursively_scan(discounts, cs, deltas):
    r""" Recursively calculate vs_minus_v_xs according to following equation:
    vs_minus_v_xs(t) = deltas(t) + discounts(t) * cs(t) * vs_minus_v_xs(t + 1)

    Args:
        discounts: A float32 tensor of shape [T, B] with discounts encountered when
                   following the behaviour policy.
        cs: A float32 tensor of shape [T, B], which corresponding to $c_s$ in the
            origin paper.
        deltas: A float32 tensor of shape [T, B], which corresponding to 
                $\delta_s * V$ in the origin paper.

    Returns:
        vs_minus_v_xs: A float32 tensor of shape [T, B], which corresponding to 
                       $v_s - V(x_s)$ in the origin paper.
    """

    # All sequences are reversed, computation starts from the back.
    reverse_discounts = layers.reverse(x=discounts, axis=[0])
    reverse_cs = layers.reverse(x=cs, axis=[0])
    reverse_deltas = layers.reverse(x=deltas, axis=[0])

    static_while = layers.StaticRNN()
    # init: shape [B]
    init = layers.fill_constant_batch_size_like(
        discounts, shape=[1], dtype='float32', value=0.0, input_dim_idx=1)

    with static_while.step():
        discount_t = static_while.step_input(reverse_discounts)
        c_t = static_while.step_input(reverse_cs)
        delta_t = static_while.step_input(reverse_deltas)

        vs_minus_v_xs_t_plus_1 = static_while.memory(init=init)
        vs_minus_v_xs_t = delta_t + discount_t * c_t * vs_minus_v_xs_t_plus_1

        static_while.update_memory(vs_minus_v_xs_t_plus_1, vs_minus_v_xs_t)

        static_while.step_output(vs_minus_v_xs_t)

    vs_minus_v_xs = static_while()

    # Reverse the results back to original order.
    vs_minus_v_xs = layers.reverse(vs_minus_v_xs, [0])

    return vs_minus_v_xs
